# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




import pylinear.array as num
import pylinear.computation as comp




def dot(x, y): 
    from operator import add
    return reduce(add, (xi*yi for xi, yi in zip(x,y)))




class StrongAdvectionOperator:
    def __init__(self, discr, a, inflow_u=None):
        self.a = a
        self.discr = discr
        self.inflow_u = inflow_u

        from hedge.flux import zero, trace_sign, make_normal, local, neighbor, average
        from hedge.discretization import bind_flux, bind_nabla, bind_mass_matrix, \
                bind_inverse_mass_matrix

        normal = make_normal(self.discr.dimensions)
        flux_weak = dot(normal, a) * average# - 0.5 *(local-neighbor)
        flux_strong = dot(normal, a)*local - flux_weak
        self.flux = bind_flux(self.discr, flux_strong)

        self.nabla = bind_nabla(discr)
        self.mass = bind_mass_matrix(discr)
        self.m_inv = bind_inverse_mass_matrix(discr)

    def rhs(self, t, u):
        from pytools import argmax
        from hedge.discretization import pair_with_boundary

        if self.inflow_u is not None:
            bc_in = self.discr.interpolate_boundary_function(
                    lambda x: self.inflow_u(t, x),
                    "inflow")

            return dot(self.a, self.nabla*u) - self.m_inv*(
                    self.flux * u + 
                    self.flux * pair_with_boundary(u, bc_in, "inflow"))
        else:
            return dot(self.a, self.nabla*u) - m_inv*(self.flux * u)





def main() :
    from hedge.element import \
            TriangularElement, \
            TetrahedralElement
    from hedge.timestep import RK4TimeStepper
    from hedge.mesh import \
            make_disk_mesh, \
            make_square_mesh, \
            make_regular_square_mesh, \
            make_single_element_mesh, \
            make_ball_mesh, \
            make_box_mesh
    from hedge.discretization import \
            Discretization, \
            bind_flux, \
            bind_nabla, \
            bind_weak_nabla, \
            bind_mass_matrix, \
            bind_inverse_mass_matrix, \
            pair_with_boundary
    from hedge.visualization import SiloVisualizer, make_silo_file
    from hedge.tools import dot
    from pytools.arithmetic_container import ArithmeticList
    from pytools.stopwatch import Job
    from math import sin, cos, pi, sqrt
    from hedge.parallel import \
            guess_parallelization_context, \
            reassemble_volume_field

    def u_analytic(t, x):
        return sin(3*(a*x+t))

    def boundary_tagger(vertices, el, face_nr):
        if el.face_normals[face_nr] * a > 0:
            return ["inflow"]
        else:
            return ["outflow"]

    pcon = guess_parallelization_context()

    dim = 2
    order = 3
    periodic = False

    if dim == 2:
        a = num.array([1,0])
        if pcon.is_head_rank:
            #mesh = make_square_mesh(boundary_tagger=boundary_tagger, max_area=0.1)
            #mesh = make_square_mesh(boundary_tagger=boundary_tagger, max_area=0.2)
            #mesh = make_regular_square_mesh(a=-r, b=r, boundary_tagger=boundary_tagger, n=3)
            #mesh = make_single_element_mesh(boundary_tagger=boundary_tagger)
            #mesh = make_disk_mesh(r=pi, boundary_tagger=boundary_tagger, max_area=0.5)
            mesh = make_disk_mesh(boundary_tagger=boundary_tagger)
        el_class = TriangularElement
    elif dim == 3:
        a = num.array([0,0,0.5])
        if pcon.is_head_rank:
            if periodic:
                mesh = make_box_mesh(dimensions=(1,1,2*pi/3),
                        periodic=periodic, max_volume=0.01)
            else:
                mesh = make_box_mesh(max_volume=0.01, 
                        boundary_tagger=boundary_tagger)
                #mesh = make_ball_mesh(boundary_tagger=boundary_tagger)
        el_class = TetrahedralElement
    else:
        raise RuntimeError, "bad number of dimensions"

    if pcon.is_head_rank:
        mesh_data = pcon.distribute_mesh(mesh)
    else:
        mesh_data = pcon.receive_mesh()

    discr = pcon.make_discretization(mesh_data, el_class(order))
    vis = SiloVisualizer(discr)
    op = StrongAdvectionOperator(discr, a, u_analytic)
    u = discr.interpolate_volume_function(lambda x: u_analytic(0, x))

    if pcon.is_head_rank:
        gdiscr = Discretization(mesh, el_class(order))
        gop = StrongAdvectionOperator(gdiscr, a, u_analytic)
        gu = gdiscr.interpolate_volume_function(lambda x: u_analytic(0, x))
        gvis = SiloVisualizer(gdiscr)
    else:
        gdiscr = None

    print "%d elements" % len(discr.mesh.elements)

    dt = discr.dt_factor(comp.norm_2(a))
    stepfactor = 1
    nsteps = int(4/dt)

    stepper = RK4TimeStepper()
    gstepper = RK4TimeStepper()
    for step in range(nsteps):
        if step % stepfactor == 0:
            print "timestep %d, t=%f, l2=%f" % (step, dt*step, sqrt(u*(op.mass*u)))

        t = step*dt
        if True:
            rhs = op.rhs(t, u)
            g_ass_u = reassemble_volume_field(pcon, gdiscr, discr, u)
            g_ass_rhs = reassemble_volume_field(pcon, gdiscr, discr, rhs)
            g_ass_bdry = reassemble_volume_field(pcon, gdiscr, discr, 
                    discr.received_bdrys[1-pcon.rank])
            if pcon.is_head_rank:
                grhs = gop.rhs(t, gu)
                silo = make_silo_file("fld-%04d" % step)
                gvis.add_to_silo(silo, [
                            ("ser_u", gu), 
                            ("par_u", g_ass_u), 
                            ("par_rhs", g_ass_rhs), 
                            ("ser_rhs", grhs), 
                            ("bdry", g_ass_bdry), 
                            ], 
                            expressions=[("d_rhs", "par_rhs-ser_rhs")],
                            time=t, 
                            step=step
                            )
                silo.close()

        u = stepper(u, t, dt, op.rhs)
        if pcon.is_head_rank:
            gu = gstepper(gu, t, dt, gop.rhs)

        #u_true = discr.interpolate_volume_function(lambda x: u_analytic(t, x))


if __name__ == "__main__":
    import cProfile as profile
    main()





"""ODE solvers: timestepping support, such as Runge-Kutta, Adams-Bashforth, etc."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




_RK4A = [0.0,
        -567301805773 /1357537059087,
        -2404267990393/2016746695238,
        -3550918686646/2091501179385,
        -1275806237668/ 842570457699,
        ]

_RK4B = [1432997174477/ 9575080441755,
        5161836677717 /13612068292357,
        1720146321549 / 2090206949498,
        3134564353537 / 4481467310338,
        2277821191437 /14882151754819,
        ]

_RK4C = [0.0,
        1432997174477/9575080441755,
        2526269341429/6820363962896,
        2006345519317/3224310063776,
        2802321613138/2924317926251,
        1,
        ]



_ABCoefficients = [
        # from R. Verfuerth,
        # "Skript Numerische Behandlung von Differentialgleichungen"
        None,
        [1],
        [3/2, -1/2],
        [23/12, -16/12, 5/12],
        [55/24, -59/24, 37/24, -9/24],
        [1901/720, -2774/720, 2616/720, -1274/720, 251/720]
        ]




class TimeStepper(object):
    pass




class RK4TimeStepper(TimeStepper):
    def __init__(self):
        from pytools.log import IntervalTimer
        self.timer = IntervalTimer(
                "t_rk4", "Time spent doing algebra in RK4")

    def add_instrumentation(self, logmgr):
        logmgr.add_quantity(self.timer)

    def __call__(self, y, t, dt, rhs):
        try:
            self.residual
        except AttributeError:
            self.residual = 0*rhs(t, y)

        for a, b, c in zip(_RK4A, _RK4B, _RK4C):
            this_rhs = rhs(t + c*dt, y)

            self.timer.start()
            self.residual = a*self.residual + dt*this_rhs
            y = y + b * self.residual
            self.timer.stop()

        return y




class AdamsBashforthTimeStepper(TimeStepper):
    def __init__(self, order, startup_stepper=RK4TimeStepper()):
        try:
            self.coefficients = _ABCoefficients[order]
        except IndexError:
            self.coefficients = None

        if self.coefficients is None:
            raise ValueError, "unsupported order in Adams-Bashforth"

        self.f_history = []
        self.startup_stepper = startup_stepper

    def __call__(self, y, t, dt, rhs):
        if len(self.f_history) == 0:
            # insert IC
            self.f_history.append(rhs(t, y))

        if len(self.f_history) < len(self.coefficients):
            ynew = self.startup_stepper(y, t, dt, rhs)
            if len(self.f_history) == len(self.coefficients) - 1:
                # here's some memory we won't need any more
                del self.startup_stepper

        else:
            from operator import add

            assert len(self.coefficients) == len(self.f_history)
            ynew = y + dt * reduce(add,
                    (coeff * f 
                        for coeff, f in 
                        zip(self.coefficients, self.f_history)))

            self.f_history.pop()

        self.f_history.insert(0, rhs(t+dt, ynew))
        return ynew




"""Jacobi polynomials and Vandermonde matrices."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""





import numpy
from pytools import memoize
import hedge._internal




JacobiFunction = hedge._internal.JacobiPolynomial
DiffJacobiFunction = hedge._internal.DiffJacobiPolynomial




class LegendreFunction(JacobiFunction):
    def __init__(self, N):
        JacobiFunction.__init__(self, 0, 0, N)
        
class DiffLegendreFunction(DiffJacobiFunction):
    def __init__(self, N):
        DiffJacobiFunction.__init__(self, 0, 0, N)




def generic_vandermonde(points, functions):
    """Return a Vandermonde matrix.

    The Vandermonde Matrix is given by M{V_{i,j} := f_j(x_i)}
    where C{functions} is the list of M{f_j} and points is 
    the list of M{x_i}.
    """
    v = numpy.zeros((len(points), len(functions)))
    for i, x in enumerate(points):
        for j, f in enumerate(functions):
            v[i,j] = f(x)
    return v




def generic_multi_vandermonde(points, functions):
    """Return multiple Vandermonde matrices.

    The Vandermonde Matrix is given by M{V_{i,j} := f_j(x_i)}
    where C{functions} is the list of M{f_j} and points is 
    the list of M{x_i}.

    The functions M{f_j} are multi-valued (i.e. return iterables), and one 
    matrix is returned for each return value.
    """
    count = len(functions[0](points[0]))
    result = [numpy.zeros((len(points), len(functions))) for n in range(count)]

    for i, x in enumerate(points):
        for j, f in enumerate(functions):
            for n, f_n in enumerate(f(x)):
                result[n][i,j] = f_n
    return result




def legendre_vandermonde(points, N):
    return generic_vandermonde(points, 
            [LegendreFunction(i) for i in range(N+1)])




# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




from __future__ import division
from hedge.tools import \
        cyl_bessel_j, \
        cyl_bessel_j_prime
from math import sqrt, pi, sin, cos, atan2
import cmath




# solution adapters -----------------------------------------------------------
class RealPartAdapter:
    def __init__(self, adaptee):
        self.adaptee = adaptee

    @property
    def shape(self):
        return self.adaptee.shape

    def __call__(self, x):
        return [xi.real for xi in self.adaptee(x)]

class SplitComplexAdapter:
    def __init__(self, adaptee):
        self.adaptee = adaptee

    @property
    def shape(self):
        (n,) = self.adaptee.shape
        return (n*2,)

    def __call__(self, x):
        ad_x = self.adaptee(x)
        return [xi.real for xi in ad_x] + [xi.imag for xi in ad_x]

class CartesianAdapter:
    def __init__(self, adaptee):
        self.adaptee = adaptee

    @property
    def shape(self):
        return self.adaptee.shape

    def __call__(self, x):
        xy = x[:2]
        r = sqrt(xy*xy)
        phi = atan2(x[1], x[0])

        prev_result = self.adaptee(x)
        result = []
        i = 0
        while i < len(prev_result):
            fr, fphi, fz = prev_result[i:i+3]
            result.extend([
                    cos(phi)*fr - sin(phi)*fphi, # ex
                    sin(phi)*fr + cos(phi)*fphi, # ey
                    fz,
                    ])
            i += 3

        return result




# actual solutions ------------------------------------------------------------
class CylindricalCavityMode:
    """A cylindrical TM cavity mode.

    Taken from:
    J.D. Jackson, Classical Electrodynamics, Wiley.
    3rd edition, 2001.
    ch 8.7, p. 368f.
    """
    
    def __init__(self,  m, n, p, radius, height, epsilon, mu):
        try:
            from bessel_zeros import bessel_zeros
        except ImportError:
            print "*** You need to generate the bessel root data file."
            print "*** Execute generate-bessel-zeros.py at the command line."
            raise

        assert m >= 0 and m == int(m)
        assert n >= 1 and n == int(n)
        assert p >= 0 and p == int(p)
        self.m = m
        self.n = n
        self.p = p
        self.phi_sign = 1

        R = self.radius = radius
        d = self.height = height

        self.epsilon = epsilon
        self.mu = mu

        self.t = 0

        x_mn = bessel_zeros[m][n-1]

        self.omega = 1 / sqrt(mu*epsilon) * sqrt(
                x_mn**2 / R**2
                + p**2 * pi**2 / d**2)

        self.gamma_mn = x_mn/R

    def set_time(self, t):
        self.t = t

    @property
    def shape(self):
        return (6,)

    def __call__(self, x):
        # coordinates -----------------------------------------------------
        xy = x[:2]
        r = sqrt(xy*xy)
        phi = atan2(x[1], x[0])
        z = x[2]

        # copy instance variables for easier access -----------------------
        m = self.m
        p = self.p
        gamma = self.gamma_mn
        phi_sign = self.phi_sign
        omega = self.omega
        d = self.height
        epsilon = self.epsilon

        # common subexpressions -------------------------------------------
        tdep = cmath.exp(-1j * omega * self.t)
        phi_factor = cmath.exp(phi_sign * 1j * m * phi)

        # psi and derivatives ---------------------------------------------
        psi = cyl_bessel_j(m, gamma * r) * phi_factor
        psi_dr = gamma*cyl_bessel_j_prime(m, gamma*r) * phi_factor
        psi_dphi = (cyl_bessel_j(m, gamma * r) 
                * 1/r * phi_sign*1j*m * phi_factor)

        # field components in polar coordinates ---------------------------
        ez   = tdep * cos(p * pi * z / d) * psi

        e_transverse_factor = (tdep
                * (-p*pi/(d*gamma**2))
                * sin(p * pi * z / d))

        er   = e_transverse_factor * psi_dr
        ephi = e_transverse_factor * psi_dphi

        hz   = 0j

        # z x grad psi = z x (psi_x, psi_y)   = (-psi_y,   psi_x)
        # z x grad psi = z x (psi_r, psi_phi) = (-psi_phi, psi_r)
        h_transverse_factor = (tdep
                * 1j*epsilon*omega/gamma**2
                * cos(p * pi * z / d))

        hr   = h_transverse_factor * (-psi_dphi)
        hphi = h_transverse_factor * psi_dr

        return [er, ephi, ez, hr, hphi, hz]




class RectangularWaveguideMode:
    """A rectangular TM cavity mode."""
    
    def __init__(self, epsilon, mu, mode_indices, 
            dimensions=(1,1,1), coefficients=(1,0,0),
            forward_coeff=1, backward_coeff=0):
        for n in mode_indices:
            assert n >= 0 and n == int(n)
        self.mode_indices = mode_indices
        self.dimensions = dimensions
        self.coefficients = coefficients
        self.forward_coeff = forward_coeff
        self.backward_coeff = backward_coeff

        self.epsilon = epsilon
        self.mu = mu

        self.t = 0

        self.factors = [n*pi/a for n,  a in zip(self.mode_indices, self.dimensions)]

        c = 1/sqrt(mu*epsilon)
        self.k = sqrt(sum(f**2 for f in self.factors))
        self.omega = self.k*c

    def set_time(self, t):
        self.t = t

    @property
    def shape(self):
        return (6,)

    def __call__(self, x):
        f,g,h = self.factors
        omega = self.omega
        k = self.k

        sx = sin(f*x[0])
        sy = sin(g*x[1])
        cx = cos(f*x[0])
        cy = cos(g*x[1])

        zdep_add = cmath.exp(1j*h*x[2])+cmath.exp(-1j*h*x[2])
        zdep_sub = cmath.exp(1j*h*x[2])-cmath.exp(-1j*h*x[2])

        tdep = cmath.exp(-1j * omega * self.t)

        C = 1j/(f**2+g**2)
        return [
                C*f*h*cx*sy*zdep_sub*tdep,
                C*g*h*sx*cy*zdep_sub*tdep,
                      sx*sy*zdep_add*tdep,
                -C*g*self.epsilon*omega*sx*cy*zdep_add*tdep,
                 C*f*self.epsilon*omega*cx*sy*zdep_add*tdep,
                0j
                ]




class RectangularCavityMode(RectangularWaveguideMode):
    """A rectangular TM cavity mode."""
    
    def __init__(self, *args, **kwargs):
        if "scale" in kwargs:
            kwargs["forward_coeff"] = scale
            kwargs["backward_coeff"] = scale
        else:
            kwargs["forward_coeff"] = 1
            kwargs["backward_coeff"] = 1
        RectangularWaveguideMode.__init__(self, *args, **kwargs)





# analytic solution tools -----------------------------------------------------
def check_time_harmonic_solution(discr, mode, c_sol):
    from hedge.discretization import bind_nabla, bind_mass_matrix
    from hedge.visualization import SiloVisualizer
    from hedge.silo import SiloFile
    from hedge.tools import dot, cross
    from hedge.silo import DB_VARTYPE_VECTOR

    def curl(field):
        return cross(nabla, field)

    vis = SiloVisualizer(discr)

    nabla = bind_nabla(discr)
    mass = bind_mass_matrix(discr)

    def rel_l2_error(err, base):
        def l2_norm(field):
            return sqrt(dot(field, mass*field))

        base_l2 = l2_norm(base)
        err_l2 = l2_norm(err)
        if base_l2 == 0:
            if err_l2 == 0:
                return 0.
            else:
                return float("inf")
        else:
            return err_l2/base_l2

    dt = 0.1

    for step in range(10):
        t = step*dt
        mode.set_time(t)
        fields = discr.interpolate_volume_function(c_sol)

        er = fields[0:3]
        hr = fields[3:6]
        ei = fields[6:9]
        hi = fields[9:12]

        silo = SiloFile("em-complex-%04d.silo" % step)
        vis.add_to_silo(silo,
                vectors=[
                    ("curl_er", curl(er)), 
                    ("om_hi", -mode.mu*mode.omega*hi), 
                    ("curl_hr", curl(hr)), 
                    ("om_ei", mode.epsilon*mode.omega*hi), 
                    ],
                expressions=[
                ("diff_er", "curl_er-om_hi", DB_VARTYPE_VECTOR),
                ("diff_hr", "curl_hr-om_ei", DB_VARTYPE_VECTOR),
                ],
                write_coarse_mesh=True,
                time=t, step=step
                )

        er_res = curl(er) + mode.mu     *mode.omega*hi
        ei_res = curl(ei) - mode.mu     *mode.omega*hr
        hr_res = curl(hr) - mode.epsilon*mode.omega*ei
        hi_res = curl(hi) + mode.epsilon*mode.omega*er

        print "time=%f, rel l2 residual in Re[E]=%g\tIm[E]=%g\tRe[H]=%g\tIm[H]=%g" % (
                t,
                rel_l2_error(er_res, er),
                rel_l2_error(ei_res, ei),
                rel_l2_error(hr_res, hr),
                rel_l2_error(hi_res, hi),
                )




import ez_setup

ez_setup.use_setuptools()

from setuptools import Extension

def setup(*args, **kwargs):
    from setuptools import setup
    import traceback
    try:
        setup(*args, **kwargs)
    except:
        traceback.print_exc()
        print "--------------------------------------------------------------------------"
        print "Sorry, your build failed. Try rerunning configure with different options."
        print "--------------------------------------------------------------------------"




# tools -----------------------------------------------------------------------
def flatten(list):
    """For an iterable of sub-iterables, generate each member of each 
    sub-iterable in turn, i.e. a flattened version of that super-iterable.

    Example: Turn [[a,b,c],[d,e,f]] into [a,b,c,d,e,f].
    """
    for sublist in list:
        for j in sublist:
            yield j




def humanize(sym_str):
    words = sym_str.lower().replace("_", " ").split(" ")
    return " ".join([word.capitalize() for word in words])




# siteconf handling -----------------------------------------------------------
def get_config():
    from setup import get_config_schema
    schema = get_config_schema()

    if not schema.have_config() and not schema.have_global_config():
        print "********************************************************"
        print "*** I have detected that you have not run configure."
        print "********************************************************"
        print "*** Additionally, no global config files were found."
        print "*** I will go ahead with the default configuration."
        print "*** In all likelihood, this will not work out."
        print "*** "
        print "*** See README_SETUP.txt for more information."
        print "*** "
        print "*** If the build does fail, just re-run configure with the"
        print "*** correct arguments, and then retry. Good luck!"
        print "********************************************************"
        print "*** HIT Ctrl-C NOW IF THIS IS NOT WHAT YOU WANT"
        print "********************************************************"

        delay = 10

        from time import sleep
        import sys
        while delay:
            sys.stdout.write("Continuing in %d seconds...   \r" % delay)
            sys.stdout.flush()
            delay -= 1
            sleep(1)

    return schema.read_config()




def hack_distutils():
    # hack distutils.sysconfig to eliminate debug flags
    # stolen from mpi4py
    import sys
    if not sys.platform.lower().startswith("win"):
        from distutils import sysconfig

        cvars = sysconfig.get_config_vars()
        cflags = cvars.get('OPT')
        if cflags:
            cflags = cflags.split()
            for bad_prefix in ('-g', '-O', '-Wstrict-prototypes'):
                for i, flag in enumerate(cflags):
                    if flag.startswith(bad_prefix):
                        cflags.pop(i)
                        break
                if flag in cflags:
                    cflags.remove(flag)
            cflags.append("-O3")
            #cflags.append("-g")
            cvars['OPT'] = str.join(' ', cflags)
            cvars["CFLAGS"] = cvars["BASECFLAGS"] + " " + cvars["OPT"]




# configure guts --------------------------------------------------------------
def default_or(a, b):
    if a is None:
        return b
    else:
        return a




class ConfigSchema:
    def __init__(self, options, conf_file="siteconf.py"):
        self.optdict = dict((opt.name, opt) for opt in options)
        self.options = options
        self.conf_file = conf_file

        from os.path import expanduser
        self.user_conf_file = expanduser("~/.aksetup-defaults.py")

        import sys
        if not sys.platform.lower().startswith("win"):
            self.global_conf_file = expanduser("/etc/aksetup-defaults.py")
        else:
            self.global_conf_file = None

    def get_default_config(self):
        return dict((opt.name, opt.default) 
                for opt in self.options)
        
    def read_config_from_pyfile(self, filename):
        result = {}
        filevars = {}
        execfile(filename, filevars)

        for key, value in filevars.iteritems():
            if key in self.optdict:
                result[key] = value

        return result

    def update_conf_file(self, filename, config):
        result = {}
        filevars = {}

        try:
            execfile(filename, filevars)
        except IOError:
            pass

        del filevars["__builtins__"]

        for key, value in config.iteritems():
            if value is not None:
                filevars[key] = value

        keys = filevars.keys()
        keys.sort()

        outf = open(filename, "w")
        for key in keys:
            outf.write("%s = %s\n" % (key, repr(filevars[key])))
        outf.close()

        return result

    def update_user_config(self, config):
        self.update_conf_file(self.user_conf_file, config)

    def update_global_config(self, config):
        if self.global_conf_file is not None:
            self.update_conf_file(self.global_conf_file, config)

    def get_default_config_with_files(self):
        result = self.get_default_config()

        import os
        
        confignames = []
        if self.global_conf_file is not None:
            confignames.append(self.global_conf_file)
        confignames.append(self.user_conf_file)

        for fn in confignames:
            if os.access(fn, os.R_OK):
                result.update(self.read_config_from_pyfile(fn))

        return result

    def have_global_config(self):
        import os
        return (os.access(self.user_conf_file, os.R_OK) or 
                os.access(self.global_conf_file, os.R_OK))

    def have_config(self):
        import os
        return os.access(self.conf_file, os.R_OK)

    def read_config(self, warn_if_none=True):
        import os
        result = self.get_default_config_with_files()
        if os.access(self.conf_file, os.R_OK):
            filevars = {}
            execfile(self.conf_file, filevars)

            for key, value in filevars.iteritems():
                if key in self.optdict:
                    result[key] = value
                elif key == "__builtins__":
                    pass
                else:
                    raise KeyError, "invalid config key in %s: %s" % (
                            self.conf_file, key)

        return result

    def add_to_configparser(self, parser, def_config=None):
        if def_config is None:
            def_config = self.get_default_config_with_files()

        for opt in self.options:
            default = default_or(def_config.get(opt.name), opt.default)
            opt.add_to_configparser(parser, default)

    def get_from_configparser(self, options):
        result = {}
        for opt in self.options:
            result[opt.name] = opt.take_from_configparser(options)
        return result

    def write_config(self, config):
        outf = open(self.conf_file, "w")
        for opt in self.options:
            value = config[opt.name]
            if value is not None:
                outf.write("%s = %s\n" % (opt.name, repr(config[opt.name])))
        outf.close()

    def make_substitutions(self, config):
        return dict((opt.name, opt.value_to_str(config[opt.name]))
                for opt in self.options)








class Option(object):
    def __init__(self, name, default=None, help=None):
        self.name = name
        self.default = default
        self.help = help

    def as_option(self):
        return self.name.lower().replace("_", "-")

    def metavar(self):
        last_underscore = self.name.rfind("_")
        return self.name[last_underscore+1:]

    def get_help(self, default):
        result = self.help
        if self.default:
            result += " (default: %s)" % self.value_to_str(
                    default_or(default, self.default))
        return result

    def value_to_str(self, default):
        return default 

    def add_to_configparser(self, parser, default=None):
        default = default_or(default, self.default)
        default_str = self.value_to_str(default)
        parser.add_option(
            "--" + self.as_option(), dest=self.name, 
            default=default,
            metavar=self.metavar(), help=self.get_help(default))

    def take_from_configparser(self, options):
        return getattr(options, self.name)

class Switch(Option):
    def add_to_configparser(self, parser, default=None):
        option = self.as_option()

        if not isinstance(self.default, bool):
            raise ValueError, "Switch options must have a default"

        if default is None:
            default = self.default

        if default:
            action = "store_false"
        else:
            action = "store_true"
            
        parser.add_option(
            "--" + self.as_option(), 
            dest=self.name, 
            help=self.get_help(default), 
            default=default,
            action=action)

class StringListOption(Option):
    def value_to_str(self, default):
        if default is None:
            return None

        return ",".join([str(el) for el in default])

    def get_help(self, default):
        return Option.get_help(self, default) + " (several ok)"

    def take_from_configparser(self, options):
        opt = getattr(options, self.name)
        if opt is None:
            return None
        else:
            if opt:
                return opt.split(",")
            else:
                return []


class IncludeDir(StringListOption):
    def __init__(self, lib_name, default=None, human_name=None, help=None):
        StringListOption.__init__(self, "%s_INC_DIR" % lib_name, default,
                help=help or ("Include directories for %s" 
                % (human_name or humanize(lib_name))))

class LibraryDir(StringListOption):
    def __init__(self, lib_name, default=None, human_name=None, help=None):
        StringListOption.__init__(self, "%s_LIB_DIR" % lib_name, default,
                help=help or ("Library directories for %s"
                % (human_name or humanize(lib_name))))

class Libraries(StringListOption):
    def __init__(self, lib_name, default=None, human_name=None, help=None):
        StringListOption.__init__(self, "%s_LIBNAME" % lib_name, default,
                help=help or ("Library names for %s (without lib or .so)" 
                % (human_name or humanize(lib_name))))







def configure_frontend():
    from optparse import OptionParser

    from setup import get_config_schema
    schema = get_config_schema()
    if schema.have_config():
        print "************************************************************"
        print "*** I have detected that you have already run configure."
        print "*** I'm taking the configured values as defaults for this"
        print "*** configure run. If you don't want this, delete the file"
        print "*** %s." % schema.conf_file
        print "************************************************************"

    import sys

    description = "generate a configuration file for this software package"
    parser = OptionParser(description=description)
    parser.add_option(
	    "--python-exe", dest="python_exe", default=sys.executable,
	    help="Which Python interpreter to use", metavar="PATH")

    parser.add_option("--prefix", default=None,
	    help="Ignored")
    parser.add_option("--enable-shared", help="Ignored", action="store_false")
    parser.add_option("--disable-static", help="Ignored", action="store_false")
    parser.add_option("--update-user", help="Update user config file (%s)" % schema.user_conf_file, 
            action="store_true")
    parser.add_option("--update-global", help="Update global config file (%s)" % schema.global_conf_file, 
            action="store_true")

    schema.add_to_configparser(parser, schema.read_config())

    options, args = parser.parse_args()

    config = schema.get_from_configparser(options)
    schema.write_config(config)

    if options.update_user:
        schema.update_user_config(config)

    if options.update_global:
        schema.update_global_config(config)

    import os
    if os.access("Makefile.in", os.F_OK):
        substs = schema.make_substitutions(config)
        substs["PYTHON_EXE"] = options.python_exe

        substitute(substs, "Makefile")




def substitute(substitutions, fname):
    import re
    var_re = re.compile(r"\$\{([A-Za-z_0-9]+)\}")
    string_var_re = re.compile(r"\$str\{([A-Za-z_0-9]+)\}")

    fname_in = fname+".in"
    lines = open(fname_in, "r").readlines()
    new_lines = []
    for l in lines:
        made_change = True
        while made_change:
            made_change = False
            match = var_re.search(l)
            if match:
                varname = match.group(1)
                l = l[:match.start()] + str(substitutions[varname]) + l[match.end():]
                made_change = True

            match = string_var_re.search(l)
            if match:
                varname = match.group(1)
                subst = substitutions[varname]
                if subst is None:
                    subst = ""
                else:
                    subst = '"%s"' % subst

                l = l[:match.start()] + subst  + l[match.end():]
                made_change = True
        new_lines.append(l)
    new_lines.insert(1, "# DO NOT EDIT THIS FILE -- it is generated by configure\n")
    import sys
    new_lines.insert(2, "# %s\n" % (" ".join(sys.argv)))
    open(fname, "w").write("".join(new_lines))

    from os import stat, chmod
    infile_stat_res = stat(fname_in)
    chmod(fname, infile_stat_res.st_mode)



# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




import numpy
import numpy.linalg as la
from hedge.tools import Rotation




def main() :
    from hedge.element import \
            TriangularElement, \
            TetrahedralElement
    from hedge.operators import StrongHeatOperator
    from hedge.timestep import RK4TimeStepper
    from hedge.mesh import \
            make_disk_mesh, \
            make_regular_square_mesh, \
            make_square_mesh, \
            make_ball_mesh
    from hedge.visualization import SiloVisualizer, VtkVisualizer
    from math import sin, cos, pi, exp, sqrt
    from hedge.parallel import guess_parallelization_context
    from hedge.data import TimeConstantGivenFunction, \
            GivenFunction, ConstantGivenFunction

    pcon = guess_parallelization_context()

    dim = 2

    def boundary_tagger(fvi, el, fn):
        if el.face_normals[fn][0] > 0:
            return ["dirichlet"]
        else:
            return ["neumann"]

    if dim == 2:
        if pcon.is_head_rank:
            mesh = make_disk_mesh(r=0.5, boundary_tagger=boundary_tagger)
            #mesh = make_regular_square_mesh(
                    #n=9, periodicity=(True,True))
            #mesh = make_regular_square_mesh(n=9)
            #mesh = make_square_mesh(max_area=0.008)
            #mesh.transform(Rotation(pi/8))
        el_class = TriangularElement
    elif dim == 3:
        if pcon.is_head_rank:
            mesh = make_ball_mesh(max_volume=0.001)
        el_class = TetrahedralElement
    else:
        raise RuntimeError, "bad number of dimensions"

    if pcon.is_head_rank:
        print "%d elements" % len(mesh.elements)
        mesh_data = pcon.distribute_mesh(mesh)
    else:
        mesh_data = pcon.receive_mesh()

    discr = pcon.make_discretization(mesh_data, el_class(3))
    stepper = RK4TimeStepper()
    vis = VtkVisualizer(discr, pcon, "fld")

    dt = discr.dt_factor(1)**2 / 5
    nsteps = int(1/dt)
    if pcon.is_head_rank:
        print "dt", dt
        print "nsteps", nsteps

    def u0(x):
        if la.norm(x) < 0.2:
            #return exp(-100*x*x)
            return 1
        else:
            return 0

    def coeff(x):
        if x[0] < 0:
            return 0.25
        else:
            return 1

    def dirichlet_bc(t, x):
        return 0

    def neumann_bc(t, x):
        return 2

    op = StrongHeatOperator(discr, 
            #coeff=coeff,
            dirichlet_tag="dirichlet",
            dirichlet_bc=TimeConstantGivenFunction(ConstantGivenFunction(0)),
            neumann_tag="neumann", 
            neumann_bc=TimeConstantGivenFunction(ConstantGivenFunction(1))
            )
    u = discr.interpolate_volume_function(u0)

    # diagnostics setup -------------------------------------------------------
    from pytools.log import LogManager, \
            add_general_quantities, \
            add_simulation_quantities, \
            add_run_info

    logmgr = LogManager("heat.dat", "w", pcon.communicator)
    add_run_info(logmgr)
    add_general_quantities(logmgr)
    add_simulation_quantities(logmgr, dt)
    discr.add_instrumentation(logmgr)

    from pytools.log import IntervalTimer
    vis_timer = IntervalTimer("t_vis", "Time spent visualizing")
    logmgr.add_quantity(vis_timer)
    stepper.add_instrumentation(logmgr)

    from hedge.log import Integral, LpNorm
    u_getter = lambda: u
    logmgr.add_quantity(LpNorm(u_getter, discr, 1, name="l1_u"))
    logmgr.add_quantity(LpNorm(u_getter, discr, name="l2_u"))

    logmgr.add_watches(["step.max", "t_sim.max", "l2_u", "t_step.max"])

    # timestep loop -----------------------------------------------------------
    for step in range(nsteps):
        logmgr.tick()
        t = step*dt

        if step % 10 == 0:
            vis_timer.start()
            visf = vis.make_file("fld-%04d" % step)
            vis.add_data(visf,
                    [("u", u), ], 
                    time=t, step=step, 
                    #write_coarse_mesh=True
                    )
            visf.close()
            vis_timer.stop()

        u = stepper(u, t, dt, op.rhs)

if __name__ == "__main__":
    #import cProfile as profile
    #profile.run("main()", "wave2d.prof")
    main()




"""Parallelism support"""

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import pytools
import numpy
import numpy.linalg as la
import hedge.discretization
import hedge.mesh




class ParallelizationContext(object):
    @property
    def rank(self):
        raise NotImplementedError

    @property
    def ranks(self):
        raise NotImplementedError

    @property
    def head_rank(self):
        raise NotImplementedError

    @property
    def is_head_rank(self):
        return self.rank == self.head_rank

    def distribute_mesh(self, mesh, partition=None):
        """Take the Mesh instance `mesh' and distribute it according to `partition'.

        If partition is an integer, invoke PyMetis to partition the mesh into this
        many parts, distributing over the first `partition' ranks.

        If partition is None, act as if partition was the integer corresponding
        to the current number of ranks on the job.

        If partition is not an integer, it must be a mapping from element number to 
        rank. (A list or tuple of rank numbers will do, for example, or so will
        a full-blown dict.)

        Returns a mesh chunk.

        We deliberately do not define the term `mesh chunk'. The return value
        of this function is to be treated as opaque by the user, only to be
        used as an argument to L{make_discretization}().

        This routine may only be invoked on the head rank.
        """
        raise NotImplementedError

    def receive_mesh(self):
        """Wait for a mesh chunk to be sent by the head rank.

        We deliberately do not define the term `mesh chunk'. The return value
        of this function is to be treated as opaque by the user, only to be
        used as an argument to L{make_discretization}().

        This routine should only be invoked on non-head ranks.
        """

        raise NotImplementedError

    def make_discretization(self, mesh_data, *args, **kwargs):
        """Construct a Discretization instance.

        `mesh_data' is whatever gets returned from distribute_mesh and
        receive_mesh(). Any extra arguments are directly forwarded to
        the respective Discretization constructor.
        """
        raise NotImplementedError





class SerialParallelizationContext(ParallelizationContext):
    communicator = None

    @property
    def rank(self):
        return 0

    @property
    def ranks(self):
        return [0]

    @property
    def head_rank(self):
        return 0

    def distribute_mesh(self, mesh, partition=None):
        return mesh

    def make_discretization(self, mesh_data, *args, **kwargs):
        from hedge.discretization import Discretization

        return Discretization(mesh_data, *args, **kwargs)




class RankData(pytools.Record):
    def __init__(self, 
            mesh, 
            global2local_elements,
            global2local_vertex_indices,
            neighbor_ranks,
            global_periodic_opposite_faces,
            old_el_numbers=None
            ):
        pytools.Record.__init__(self, locals())

    def reordered_by(self, *args, **kwargs):
        old_el_numbers = self.mesh.get_reorder_oldnumbers(*args, **kwargs)
        mesh = self.mesh.reordered(old_el_numbers)
        return self.copy(
                mesh=mesh,
                old_el_numbers=old_el_numbers
                )




class MPIParallelizationContext(ParallelizationContext):
    def __init__(self, communicator):
        self.communicator = communicator

    @property
    def rank(self):
        return self.communicator.rank

    @property
    def ranks(self):
        return range(0, self.communicator.size)

    @property
    def head_rank(self):
        return 0

    def distribute_mesh(self, mesh, partition=None):
        assert self.is_head_rank

        if partition is None:
            partition = len(self.ranks)

        # compute partition using Metis, if necessary
        if isinstance(partition, int):
            if partition == 1:
                return 

            from pymetis import part_graph
            cuts, partition = part_graph(partition, 
                    mesh.element_adjacency_graph())

        # find ranks to which we need to distribute
        target_ranks = set()
        for el in mesh.elements:
            target_ranks.add(partition[el.id])
        target_ranks = list(target_ranks)

        # prepare a mapping of elements to tags to speed up
        # copy_el_tagger, below
        el2tags = {}
        for tag, elements in mesh.tag_to_elements.iteritems():
            if tag == hedge.mesh.TAG_ALL:
                continue
            for el in elements:
                el2tags.setdefault(el, []).append(tag)

        # prepare a mapping of (el, face_nr) to boundary_tags
        # to speed up parallelizer_bdry_tagger, below
        elface2tags = {}
        for tag, elfaces in mesh.tag_to_boundary.iteritems():
            if tag == hedge.mesh.TAG_ALL:
                continue
            for el, fn in elfaces:
                elface2tags.setdefault((el, fn), []).append(tag)

        # prepare a mapping from (el, face_nr) to the rank
        # at the other end of the interface, if different from
        # current. concurrently, prepare a mapping 
        #  rank -> set([ranks that border me])
        elface2rank = {}
        neighboring_ranks = {}

        for elface1, elface2 in mesh.interfaces:
            e1, f1 = elface1
            e2, f2 = elface2
            r1 = partition[e1.id]
            r2 = partition[e2.id]

            if r1 != r2:
                neighboring_ranks.setdefault(r1, set()).add(r2)
                neighboring_ranks.setdefault(r2, set()).add(r1)

                elface2rank[elface1] = r2
                elface2rank[elface2] = r1

        # prepare a new mesh for each rank and send it
        import boost.mpi as mpi
        from hedge.mesh import TAG_NO_BOUNDARY

        for rank in target_ranks:
            rank_global_elements = [el 
                    for el in mesh.elements
                    if partition [el.id] == rank]

            # pick out this rank's vertices
            from pytools import flatten
            rank_global_vertex_indices = set(flatten(
                    el.vertex_indices for el in rank_global_elements))

            rank_local_vertices = [mesh.points[vi] 
                    for vi in rank_global_vertex_indices]

            # find global-to-local maps
            rank_global2local_vertex_indices = dict(
                    (gvi, lvi) for lvi, gvi in 
                    enumerate(rank_global_vertex_indices))

            rank_global2local_elements = dict(
                    (el.id, i) for i, el in 
                    enumerate(rank_global_elements))

            # find elements in local numbering
            rank_local_elements = [
                    [rank_global2local_vertex_indices[vi] 
                        for vi in el.vertex_indices]
                    for el in rank_global_elements]

            # make new local Mesh object, including 
            # boundary and element tagging
            def parallelizer_bdry_tagger(fvi, local_el, fn):
                el = rank_global_elements[local_el.id]

                result = elface2tags.get((el, fn), [])
                try:
                    opp_rank = elface2rank[el, fn]
                    result.append(hedge.mesh.TAG_RANK_BOUNDARY(opp_rank))

                    # keeps this part of the boundary from falling
                    # under the "None" tag.
                    result.append(TAG_NO_BOUNDARY)
                except KeyError:
                    pass

                return result

            def copy_el_tagger(local_el):
                return el2tags.get(rank_global_elements[local_el.id], [])

            def is_rankbdry_face((local_el, face_nr)):
                return (rank_global_elements[local_el.id], face_nr) in elface2rank

            from hedge.mesh import make_conformal_mesh
            rank_mesh = make_conformal_mesh(
                    rank_local_vertices,
                    rank_local_elements,
                    parallelizer_bdry_tagger, copy_el_tagger,
                    mesh.periodicity,
                    is_rankbdry_face)

            # assemble per-rank data
            rank_data = RankData(
                    rank_mesh, 
                    rank_global2local_elements,
                    rank_global2local_vertex_indices,
                    neighboring_ranks[rank],
                    mesh.periodic_opposite_faces)

            if rank == self.head_rank:
                result = rank_data
            else:
                self.communicator.send(rank, 0, rank_data)

        return result

    def receive_mesh(self):
        return self.communicator.recv(self.head_rank, 0)

    def make_discretization(self, mesh_data, *args, **kwargs):
        return ParallelDiscretization(self, mesh_data, *args, **kwargs)




class ParallelDiscretization(hedge.discretization.Discretization):
    def __init__(self, pcon, rank_data, local_discretization, debug=False):
        self.received_bdrys = {}
        self.context = pcon

        hedge.discretization.Discretization.__init__(self,
                rank_data.mesh, local_discretization, debug=debug)

        self.global2local_vertex_indices = rank_data.global2local_vertex_indices 
        self.neighbor_ranks = rank_data.neighbor_ranks
        self.global_periodic_opposite_faces = rank_data.global_periodic_opposite_faces

        if rank_data.old_el_numbers is not None:
            from hedge.tools import reverse_lookup_table
            new_el_numbers = reverse_lookup_table(rank_data.old_el_numbers)
            self.global2local_elements = dict(
                    (gi, new_el_numbers[li])
                    for gi, li in rank_data.global2local_elements.iteritems())
        else:
            self.global2local_elements = rank_data.global2local_elements 

        self._setup_neighbor_connections()

        # instrumentation -----------------------------------------------------
        from pytools.log import IntervalTimer, EventCounter

        self.comm_flux_counter = EventCounter("n_comm_flux", 
                "Number of inner flux communication runs")
        self.comm_flux_timer = IntervalTimer("t_comm_flux", 
                "Time spent communicating to compute inner fluxes")

    def add_instrumentation(self, mgr):
        mgr.add_quantity(self.comm_flux_counter)
        mgr.add_quantity(self.comm_flux_timer)

        hedge.discretization.Discretization.add_instrumentation(self, mgr)

    def _setup_neighbor_connections(self):
        import boost.mpi as mpi

        comm = self.context.communicator

        def get_flux_face(el_face):
            face_group, ffi = (rank_discr_boundary
                    .el_face_to_face_group_and_flux_face_index[el_face])
            return face_group.flux_faces[ffi]

        if self.neighbor_ranks:
            # send interface information to neighboring ranks -----------------
            from pytools import reverse_dictionary
            local2global_vertex_indices = \
                    reverse_dictionary(self.global2local_vertex_indices)

            send_requests = mpi.RequestList()

            for rank in self.neighbor_ranks:
                bdry_tag = hedge.mesh.TAG_RANK_BOUNDARY(rank)
                rank_bdry = self.mesh.tag_to_boundary[bdry_tag]
                rank_discr_boundary = self._get_boundary(bdry_tag)

                # a list of global vertex numbers for each face
                my_vertices_global = [
                        tuple(local2global_vertex_indices[vi]
                            for vi in el.faces[face_nr])
                        for el, face_nr in rank_bdry]

                # a list of node coordinates, indicating the order
                # in which nodal values will be sent, this is for
                # testing only and could (potentially) be omitted

                my_node_coords = []
                for el, face_nr in rank_bdry:
                    (estart, eend), ldis = self.find_el_data(el.id)
                    findices = ldis.face_indices()[face_nr]
                    my_node_coords.append(
                            [self.nodes[estart+i] for i in findices])

                # compile a list of FluxFace.h values for unification
                # across the rank boundary

                my_h_values = [get_flux_face(el_face).h for el_face in rank_bdry]
                
                packet = (my_vertices_global, my_node_coords, my_h_values)

                send_requests.append(comm.isend(rank, 0, packet))

            received_packets = {}
            while len(received_packets) < len(self.neighbor_ranks):
                received_packet, status = comm.recv(tag=0, return_status=True)
                received_packets[status.source] = received_packet

            mpi.wait_all(send_requests)

            # process received packets ----------------------------------------
            from pytools import flatten

            # nb_ stands for neighbor_

            self.from_neighbor_maps = {}

            for rank, (nb_all_facevertices_global, nb_node_coords, nb_h_values) in \
                    received_packets.iteritems():
                bdry_tag = hedge.mesh.TAG_RANK_BOUNDARY(rank)
                rank_bdry = self.mesh.tag_to_boundary[bdry_tag]
                rank_discr_boundary = self._get_boundary(bdry_tag)

                flat_nb_node_coords = list(flatten(nb_node_coords))

                # step 1: find start node indices for each 
                # of the neighbor's elements
                nb_face_starts = [0]
                for node_coords in nb_node_coords[:-1]:
                    nb_face_starts.append(
                            nb_face_starts[-1]+len(node_coords))

                # step 2: match faces by matching vertices
                nb_face_order = dict(
                        (frozenset(vertices), i)
                        for i, vertices in enumerate(nb_all_facevertices_global))

                # step 3: make a list of indices into the data we
                # receive from our neighbor that'll tell us how
                # to reshuffle them to match our node order
                from_indices = []

                shuffled_indices_cache = {}

                def get_shuffled_indices(face_node_count, shuffle_op):
                    try:
                        return shuffled_indices_cache[shuffle_op]
                    except KeyError:
                        unshuffled_indices = range(face_node_count)
                        result = shuffled_indices_cache[shuffle_op] = \
                                shuffle_op(unshuffled_indices)
                        return result

                for el, face_nr in rank_bdry:
                    (estart, eend), ldis = self.find_el_data(el.id)

                    my_vertices = el.faces[face_nr]
                    my_global_vertices = tuple(local2global_vertex_indices[vi]
                            for vi in my_vertices)

                    try:
                        nb_face_idx = nb_face_order[frozenset(my_global_vertices)]
                        # continue below in else part
                    except KeyError:
                        # ok, our face must be part of a periodic pair
                        my_vertices_there, axis = self.global_periodic_opposite_faces[
                                my_global_vertices]
                        nb_face_idx = nb_face_order[frozenset(my_vertices_there)]

                        his_vertices_here, axis2 = self.global_periodic_opposite_faces[
                                nb_all_facevertices_global[nb_face_idx]]

                        assert axis == axis2
                        
                        face_node_count = len(nb_node_coords[nb_face_idx])
                        nb_face_start = nb_face_starts[nb_face_idx]

                        shuffle_op = \
                                ldis.get_face_index_shuffle_to_match(
                                        my_global_vertices,
                                        his_vertices_here) 

                        shuffled_other_node_indices = [nb_face_start+i 
                                for i in get_shuffled_indices(
                                    face_node_count, shuffle_op)]

                        from_indices.extend(shuffled_other_node_indices)

                        # check if the nodes really match up
                        if self.debug:
                            my_node_indices = [estart+i for i in ldis.face_indices()[face_nr]]

                            for my_i, other_i in zip(my_node_indices, shuffled_other_node_indices):
                                dist = self.nodes[my_i]-flat_nb_node_coords[other_i]
                                dist[axis] = 0
                                assert la.norm(dist) < 1e-14
                    else:
                        # continue handling of nonperiodic case
                        nb_vertices = nb_all_facevertices_global[nb_face_idx]

                        face_node_count = len(nb_node_coords[nb_face_idx])
                        nb_face_start = nb_face_starts[nb_face_idx]

                        shuffle_op = \
                                ldis.get_face_index_shuffle_to_match(
                                        my_global_vertices,
                                        nb_vertices)

                        shuffled_other_node_indices = [nb_face_start+i 
                                for i in get_shuffled_indices(
                                    face_node_count, shuffle_op)]

                        from_indices.extend(shuffled_other_node_indices)

                        # check if the nodes really match up
                        if self.debug:
                            my_node_indices = [estart+i for i in ldis.face_indices()[face_nr]]

                            for my_i, other_i in zip(my_node_indices, shuffled_other_node_indices):
                                dist = self.nodes[my_i]-flat_nb_node_coords[other_i]
                                assert la.norm(dist) < 1e-14

                    # finally, unify FluxFace.h values across boundary
                    nb_h = nb_h_values[nb_face_idx]
                    flux_face = get_flux_face((el, face_nr))
                    flux_face.h = max(nb_h, flux_face.h)

                if self.debug:
                    assert len(from_indices) == len(flat_nb_node_coords)

                # turn from_indices into an IndexMap

                from hedge._internal import IndexMap
                self.from_neighbor_maps[rank] = IndexMap(
                        len(from_indices), len(from_indices), from_indices)

    def dt_non_geometric_factor(self):
        import boost.mpi as mpi
        from hedge.discretization import Discretization
        return mpi.all_reduce(self.context.communicator, 
                Discretization.dt_non_geometric_factor(self),
                min)

    def dt_geometric_factor(self):
        import boost.mpi as mpi
        from hedge.discretization import Discretization
        return mpi.all_reduce(self.context.communicator, 
                Discretization.dt_geometric_factor(self),
                min)

    def get_flux_operator(self, flux, direct=True):
        """Return a flux operator that can be multiplied with
        a volume field to obtain the lifted interior fluxes
        or with a boundary pair to obtain the lifted boundary
        flux.

        `direct' determines whether the operator is applied in a
        matrix-free fashion or uses precomputed matrices.
        """
        return _ParallelFluxOperator(self, flux, direct)




class _ParallelFluxOperator(object):
    def __init__(self, discr, flux, direct):
        self.discr = discr
        from hedge.discretization import Discretization
        self.serial_flux_op = Discretization.get_flux_operator(
                discr, flux, direct)

    def __mul__(self, field):
        from hedge.discretization import pair_with_boundary, BoundaryPair

        if isinstance(field, BoundaryPair):
            return self.serial_flux_op * field

        import boost.mpi as mpi
        from hedge._internal import irecv_vector, isend_vector

        self.discr.comm_flux_counter.add()
        self.discr.comm_flux_timer.start()

        comm = self.discr.context.communicator

        from hedge.tools import log_shape
        ls = log_shape(field)

        if ls != ():
            assert len(ls) == 1
            field_count = ls[0]
        else:
            field_count = None

        def fix_shape(field, tag):
            if field_count is not None:
                bdry_len = len(self.discr._get_boundary(tag).nodes)
                return numpy.reshape(field, (field_count, bdry_len))
            else:
                return field

        # Subtlety here: The vectors for isend and irecv need to stay allocated
        # for as long as the request is not completed. The wrapper aids this
        # by making sure the vector outlives the request by using Boost.Python's
        # with_custodian_and_ward mechanism. However, this "life support" gets
        # eliminated if the only reference to the request is from inside a 
        # RequestList, so we need to provide our own life support for these vectors.

        neigh_recv_vecs = dict(
                (rank, self.discr.boundary_zeros(
                    hedge.mesh.TAG_RANK_BOUNDARY(rank), 
                    shape=ls))
                for rank in self.discr.neighbor_ranks)

        recv_requests = mpi.RequestList(
                irecv_vector(comm, rank, 1, neigh_recv_vecs[rank])
                for rank in self.discr.neighbor_ranks)

        send_requests = [isend_vector(comm, rank, 1,
            self.discr.boundarize_volume_field(
                field, 
                hedge.mesh.TAG_RANK_BOUNDARY(rank))
            )
            for rank in self.discr.neighbor_ranks]

        result = [self.serial_flux_op * field]

        def receive(_, status):
            from hedge.tools import apply_index_map

            tag = hedge.mesh.TAG_RANK_BOUNDARY(status.source)

            foreign_order_bfield = fix_shape(neigh_recv_vecs[status.source], tag)

            bfield = apply_index_map(
                    self.discr.from_neighbor_maps[status.source],
                    foreign_order_bfield)

            result[0] += self.serial_flux_op * \
                    pair_with_boundary(field, bfield, tag)

        mpi.wait_all(recv_requests, receive)
        mpi.wait_all(mpi.RequestList(send_requests))

        self.discr.comm_flux_timer.stop()

        return result[0]





def guess_parallelization_context():
    try:
        import boost.mpi as mpi

        if mpi.size == 1:
            return SerialParallelizationContext()
        else:
            return MPIParallelizationContext(mpi.world)
    except ImportError:
        return SerialParallelizationContext()




def reassemble_volume_field(pcon, global_discr, local_discr, field):
    from pytools import reverse_dictionary
    local2global_element = reverse_dictionary(
            local_discr.global2local_elements)

    send_packet = {}
    for eg in local_discr.element_groups:
        for el, (estart, eend) in zip(eg.members, eg.ranges):
            send_packet[local2global_element[el.id]] = field[estart:eend]

    def reduction(a, b):
        a.update(b)
        return a

    import boost.mpi as mpi

    gfield_parts = mpi.reduce(
            pcon.communicator, send_packet, reduction, pcon.head_rank)

    if pcon.is_head_rank:
        result = global_discr.volume_zeros()
        for eg in global_discr.element_groups:
            for el, (estart, eend) in zip(eg.members, eg.ranges):
                my_part = gfield_parts[el.id]
                assert len(my_part) == eend-estart
                result[estart:eend] = my_part
        return result
    else:
        return None



from __future__ import division

import pylinear.array as num
import pylinear.computation as comp
import hedge.tools
import hedge.mesh
import hedge.data




class Diagonalized1DWaveOperator:
    def __init__(self, c, discr, source_f=None, 
            flux_type="upwind",
            dirichlet_tag=hedge.mesh.TAG_ALL,
            neumann_tag=hedge.mesh.TAG_NONE,
            radiation_tag=hedge.mesh.TAG_NONE):
        self.c = c
        self.discr = discr
        self.source_f = source_f

        assert c > 0, "wave speed has to be positive"

        self.dirichlet_tag = dirichlet_tag
        self.neumann_tag = neumann_tag
        self.radiation_tag = radiation_tag

        from hedge.mesh import check_bc_coverage
        check_bc_coverage(discr.mesh, [
            dirichlet_tag,
            neumann_tag,
            radiation_tag])

        from hedge.flux import FluxVectorPlaceholder, make_normal

        dim = discr.dimensions
        s = FluxVectorPlaceholder(2)
        normal = make_normal(dim)

        from pytools.arithmetic_container import join_fields
        from hedge.tools import dot

        coeff_sign = self.coeff_sign = join_fields(-1, 1)
        flux_weak = s.avg*normal[0]*coeff_sign
        flux_strong = coeff_sign * s.int * normal[0] - flux_weak

        self.flux = discr.get_flux_operator(self.c*flux_strong)

        self.nabla = discr.nabla
        self.mass = discr.mass_operator
        self.m_inv = discr.inverse_mass_operator

        self.radiation_normals = discr.boundary_normals(self.radiation_tag)
        from pytools.arithmetic_container import outer_product, ArithmeticListMatrix
        self.radn_outer_radn = outer_product(
                self.radiation_normals, 
                self.radiation_normals,
                mult_op=num.multiply)

        from math import sqrt
        v = num.array([[1,1],[1,-1]])/sqrt(2)
        self.V = ArithmeticListMatrix(v)
        self.Vt = ArithmeticListMatrix(v.T)

    def rhs(self, t, w):
        from hedge.discretization import pair_with_boundary, cache_diff_results
        from pytools.arithmetic_container import join_fields
        from hedge.tools import dot

        s = self.Vt*w

        from pytools.arithmetic_container import work_with_arithmetic_containers
        ac_multiply = work_with_arithmetic_containers(num.multiply)

        rad_s = self.discr.boundarize_volume_field(s, self.radiation_tag)
        rad_n = self.radiation_normals
        ind_right = (rad_n[0]+1)/2
        ind_left = -(rad_n[0]-1)/2
        rad_bc = join_fields(
                num.multiply(ind_left, rad_s[0]),
                num.multiply(ind_right, rad_s[1])
                )
        self.rad_bc = rad_bc
        rhs = (-self.c*self.coeff_sign*(self.nabla[0]*cache_diff_results(s))
                + self.m_inv * (
                self.flux*s 
                + self.flux * pair_with_boundary(s, rad_bc, self.radiation_tag)
                ))

        w_rhs = self.V*rhs 
        if self.source_f is not None:
            w_rhs[0] += self.source_f(t)

        return w_rhs

    def max_eigenvalue(self):
        return self.c








"""Miscellaneous helper facilities."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""





import numpy
import numpy.linalg as la
import pyublas
import hedge._internal
from pytools.arithmetic_container import work_with_arithmetic_containers



cyl_bessel_j = hedge._internal.cyl_bessel_j
cyl_neumann = hedge._internal.cyl_neumann




def cyl_bessel_j_prime(nu, z):
    if nu == 0:
        if z == 0:
            return 0
        else:
            return -cyl_bessel_j(nu+1, z)+nu/z*cyl_bessel_j(nu, z)
    else:
        return 0.5*(cyl_bessel_j(nu-1, z)-cyl_bessel_j(nu+1, z))




AffineMap = hedge._internal.AffineMap
def _affine_map_jacobian(self):
    try:
        return self._jacobian
    except AttributeError:
        self._jacobian = numpy.linalg.det(self.matrix)
        return self._jacobian
AffineMap.jacobian = property(_affine_map_jacobian)

def _affine_map_inverted(self):
    """Return a new AffineMap that is the inverse of this one.
    """
    return AffineMap(numpy.asarray(numpy.linalg.inv(self.matrix), order="c"), 
            -numpy.linalg.solve(self.matrix, self.vector))
AffineMap.inverted = _affine_map_inverted

def _affine_map___getinitargs__(self):
    return self.matrix, self.vector
AffineMap.__getinitargs__ = _affine_map___getinitargs__




class Rotation(AffineMap):
    def __init__(self, angle):
        # FIXME: Add axis, make multidimensional
        from math import sin, cos
        AffineMap.__init__(self,
                num.array([
                    [cos(angle), sin(angle)],
                    [-sin(angle), cos(angle)]]),
                num.zeros((2,)))




class Reflection(AffineMap):
    def __init__(self, axis, dim):
        mat = num.identity(dim)
        mat[axis,axis] = -1
        AffineMap.__init__(self, mat, num.zeros((dim,)))




def plot_1d(f, a, b, steps=100, driver=None):
    h = (b - a)/steps

    points = []
    data = []
    for n in range(steps):
        x = a + h * n
        points.append(x)
        data.append(f(x))

    # autodetect driver
    if driver is None:
        try:
            import pylab
            driver = "matplotlib"
        except ImportError:
            pass
    if driver is None:
        try:
            import Gnuplot
            driver = "gnuplot"
        except ImportError:
            pass

    # actually plot
    if driver == "matplotlib":
        from pylab import plot, show
        plot(points, data)
        show()
    elif driver == "gnuplot":
        from Gnuplot import Gnuplot, Data
        gp = Gnuplot()
        gp.plot(Data(points, data))
        raw_input()
    else:
        raise ValueError, "invalid plot driver '%s'" % driver




# obj array helpers -----------------------------------------------------------
def is_obj_array(val):
    try:
        return val.dtype == object
    except AttributeError:
        return False




def to_obj_array(ary):
    ls = log_shape(ary)
    result = numpy.empty(ls, dtype=object)

    from pytools import indices_in_shape
    for i in indices_in_shape(ls):
        result[i] = ary[i]

    return result




def make_obj_array(res_list):
    result = numpy.empty((len(res_list),), dtype=object)
    for i, v in enumerate(res_list):
        result[i] = v

    return result




def join_fields(*args):
    res_list = []
    for arg in args:
        if isinstance(arg, list):
            res_list.extend(arg)
        elif isinstance(arg, numpy.ndarray):
            if log_shape(arg) == ():
                res_list.append(arg)
            else:
                res_list.extend(arg)
        else:
            res_list.append(arg)

    return make_obj_array(res_list)




def log_shape(array):
    """Returns the "logical shape" of the array.

    The "logical shape" is the shape that's left when the node-depending
    dimension has been eliminated."""

    try:
        if array.dtype.char == "O":
            return array.shape
        else:
            return array.shape[:-1]
    except AttributeError:
        return ()




def ptwise_dot(a1, a2):
    if a1.dtype == object:
        a1_log_shape = a1.shape
        a2_log_shape = log_shape(a2)

        assert a1_log_shape[-1] == a2_log_shape[0]
        len_k = a2_log_shape[0]

        result = numpy.empty(a1_log_shape[:-1]+a2_log_shape[1:], dtype=a1.dtype)
    else:
        dofs = a2.shape[-1]
        a2_log_shape = a2.shape[:-1]

        if a1.shape[-1] == dofs:
            a1_log_shape = a1.shape[:-1]
        else:
            a1_log_shape = a1.shape

        assert a1_log_shape[-1] == a2_log_shape[0]
        len_k = a2_log_shape[0]

        result = numpy.empty(a1_log_shape[:-1]+a2_log_shape[1:]+(dofs,), dtype=a1.dtype)

    from pytools import indices_in_shape
    for a1_i in indices_in_shape(a1_log_shape[:-1]):
        for a2_i in indices_in_shape(a2_log_shape[1:]):
            result[a1_i+a2_i] = sum(
                    a1[a1_i+(k,)] * a2[(k,)+a2_i]
                    for k in xrange(len_k)
                    )

    return result




def amap(f, obj_array):
    return numpy.array([f(x) for x in obj_array], dtype=object)




def levi_civita(tuple):
    """Compute an entry of the Levi-Civita tensor for the indices C{tuple}.

    Only three-tuples are supported for now.
    """
    if len(tuple) == 3:
        if tuple in [(0,1,2), (2,0,1), (1,2,0)]: 
            return 1
        elif tuple in [(2,1,0), (0,2,1), (1,0,2)]: 
            return -1
        else:
            return 0
    else:
        raise NotImplementedError




class SubsettableCrossProduct:
    """A cross product that can operate on an arbitrary subsets of its
    two operands and return an arbitrary subset of its result.
    """

    full_subset = (True, True, True)

    def __init__(self, op1_subset=full_subset, op2_subset=full_subset, result_subset=full_subset):
        """Construct a subset-able cross product.

        @arg op1_subset: The subset of indices of operand 1 to be taken into account.
          Given as a 3-sequence of bools.
        @arg op2_subset: The subset of indices of operand 2 to be taken into account.
          Given as a 3-sequence of bools.
        @arg result_subset: The subset of indices of the result that are calculated.
          Given as a 3-sequence of bools.
        """
        def subset_indices(subset):
            return [i for i, use_component in enumerate(subset) 
                    if use_component]

        self.op1_subset = op1_subset
        self.op2_subset = op2_subset
        self.result_subset = result_subset

        import pymbolic
        op1 = pymbolic.var("x")
        op2 = pymbolic.var("y")

        self.functions = []
        self.component_lcjk = []
        for i, use_component in enumerate(result_subset):
            if use_component:
                this_expr = 0
                this_component = []
                for j, j_real in enumerate(subset_indices(op1_subset)):
                    for k, k_real in enumerate(subset_indices(op2_subset)):
                        lc = levi_civita((i, j_real, k_real))
                        if lc != 0:
                            this_expr += lc*op1[j]*op2[k]
                            this_component.append((lc, j, k))
                self.functions.append(pymbolic.compile(this_expr))
                self.component_lcjk.append(this_component)

    def __call__(self, x, y, three_mult=None):
        """Compute the subsetted cross product on the indexables C{x} and C{y}.

        @arg three_mult: a function of three arguments C{sign, xj, yk}
          used in place of the product C{sign*xj*yk}. Defaults to just this
          product if not given.
        """
        if three_mult is None:
            return join_fields(*[f(x, y) for f in self.functions])
        else:
            return join_fields(
                    *[sum(three_mult(lc, x[j], y[k]) for lc, j, k in lcjk)
                    for lcjk in self.component_lcjk])




cross = SubsettableCrossProduct()




def normalize(v):
    return v/numpy.linalg.norm(v)




def sign(x):
    if x > 0: 
        return 1
    elif x == 0:
        return 0
    else: 
        return -1




def find_matching_vertices_along_axis(axis, points_a, points_b, numbers_a, numbers_b):
    a_to_b = {}
    not_found = []

    for i, pi in enumerate(points_a):
        found = False
        for j, pj in enumerate(points_b):
            dist = pi-pj
            dist[axis] = 0
            if la.norm(dist) < 1e-12:
                a_to_b[numbers_a[i]] = numbers_b[j]
                found = True
                break
        if not found:
            not_found.append(numbers_a[i])

    return a_to_b, not_found




make_vector_target = hedge._internal.make_vector_target




# linear algebra tools --------------------------------------------------------
def orthonormalize(vectors, discard_threshold=None):
    """Carry out a modified [1] Gram-Schmidt orthonormalization on
    vectors.

    If, during orthonormalization, the 2-norm of a vector drops 
    below C{discard_threshold}, then this vector is silently 
    discarded. If C{discard_threshold} is C{None}, then no vector
    will ever be dropped, and a zero 2-norm encountered during
    orthonormalization will throw an L{OrthonormalizationError}.

    [1] U{http://en.wikipedia.org/wiki/Gram%E2%80%93Schmidt_process}
    """

    from numpy import dot
    done_vectors = []

    for v in vectors:
        my_v = v.copy()
        for done_v in done_vectors:
            my_v = my_v - dot(my_v, done_v.conjugate()) * done_v
        v_norm = la.norm(my_v)

        if discard_threshold is None:
            if v_norm == 0:
                raise RuntimeError, "Orthogonalization failed"
        else:
            if v_norm < discard_threshold:
                continue

        my_v /= v_norm
        done_vectors.append(my_v)

    return done_vectors




def permutation_matrix(to_indices=None, from_indices=None, h=None, w=None,
        dtype=None, flavor=None):
    """Return a permutation matrix.

    If to_indices is specified, the resulting permutation 
    matrix P satisfies the condition

    P * e[i] = e[to_indices[i]] for i=1,...,len(to_indices)

    where e[i] is the i-th unit vector. The height of P is 
    determined either implicitly by the maximum of to_indices
    or explicitly by the parameter h.

    If from_indices is specified, the resulting permutation 
    matrix P satisfies the condition

    P * e[from_indices[i]] = e[i] for i=1,...,len(from_indices)
    
    where e[i] is the i-th unit vector. The width of P is
    determined either implicitly by the maximum of from_indices
    of explicitly by the parameter w.

    If both to_indices and from_indices is specified, a ValueError
    exception is raised.
    """
    if to_indices is not None and from_indices is not None:
        raise ValueError, "only one of to_indices and from_indices may " \
                "be specified"

    if to_indices is not None:
        if h is None:
            h = max(to_indices)+1
        w = len(to_indices)
    else:
        if w is None:
            w = max(from_indices)+1
        h = len(from_indices)

    if flavor is None:
        result = numpy.zeros((h,w), dtype=dtype)

        if to_indices is not None:
            for j, i in enumerate(to_indices):
                result[i,j] = 1
        else:
            for i, j in enumerate(from_indices):
                result[i,j] = 1
    else:
        result = pyublas.zeros((h,w), dtype=dtype, flavor=flavor)

        if to_indices is not None:
            for j, i in enumerate(to_indices):
                result.add_element(i, j, 1)
        else:
            for i, j in enumerate(from_indices):
                result.add_element(i, j, 1)

    return result




def leftsolve(A, B):
    return la.solve(A.T, B.T).T




def unit_vector(n, i, dtype=None):
    """Return the i-th unit vector of size n, with the given dtype."""
    result = numpy.zeros((n,), dtype=dtype)
    result[i] = 1
    return result




# eoc estimation --------------------------------------------------------------
def estimate_order_of_convergence(abscissae, errors):
    """Assuming that abscissae and errors are connected by a law of the form

    error = constant * abscissa ^ (-order),

    this function finds, in a least-squares sense, the best approximation of
    constant and order for the given data set. It returns a tuple (constant, order).
    Both inputs must be PyLinear vectors.
    """
    assert len(abscissae) == len(errors)
    if len(abscissae) <= 1:
        raise RuntimeError, "Need more than one value to guess order of convergence."

    coefficients = numpy.polyfit(numpy.log10(abscissae), numpy.log10(errors), 1)
    return 10**coefficients[-1], -coefficients[-2]


  

class EOCRecorder(object):
    def __init__(self):
        self.history = []

    def add_data_point(self, abscissa, error):
        self.history.append((abscissa, error))

    def estimate_order_of_convergence(self, gliding_mean = None):
        abscissae = numpy.array([ a for a,e in self.history ])
        errors = numpy.array([ e for a,e in self.history ])

        size = len(abscissae)
        if gliding_mean is None:
            gliding_mean = size

        data_points = size - gliding_mean + 1
        result = numpy.zeros((data_points, 2), float)
        for i in range(data_points):
            result[i,0], result[i,1] = estimate_order_of_convergence(
                abscissae[i:i+gliding_mean], errors[i:i+gliding_mean])
        return result

    def pretty_print(self, abscissa_label="N", error_label="Error", gliding_mean=2):
        from pytools import Table

        tbl = Table()
        tbl.add_row((abscissa_label, error_label, "Running EOC"))

        gm_eoc = self.estimate_order_of_convergence(gliding_mean)
        for i, (absc, err) in enumerate(self.history):
            if i < gliding_mean-1:
                tbl.add_row((str(absc), str(err), ""))
            else:
                tbl.add_row((str(absc), str(err), str(gm_eoc[i-gliding_mean+1,1])))

        if len(self.history) > 1:
            return str(tbl) + "\n\nOverall EOC: %s" % self.estimate_order_of_convergence()[0,1]
        else:
            return str(tbl)

    def write_gnuplot_file(self, filename):
        outfile = file(filename, "w")
        for absc, err in self.history:
            outfile.write("%f %f\n" % (absc, err))
        result = self.estimate_order_of_convergence()
        const = result[0,0]
        order = result[0,1]
        outfile.write("\n")
        for absc, err in self.history:
            outfile.write("%f %f\n" % (absc, const * absc**(-order)))




# small utilities -------------------------------------------------------------
class Closable(object):
    def __init__(self):
        self.is_closed = False

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        if not self.is_closed:
            # even if close attempt fails, consider ourselves closed still
            try:
                self.do_close()
            finally:
                self.is_closed = True




def mem_checkpoint(name=None):
    """Invoke the garbage collector and wait for a keypress."""

    import gc
    gc.collect()
    if name:
        raw_input("%s -- hit Enter:" % name)
    else:
        raw_input("Enter:")





class DisabledFixedSizeSliceAdapter(object):
    """Adapts an indexable object C{idxable} so that C{idxable[i]}
    refers to the slice C{adaptee[i*unit:(i+1)*unit]}. This effectively
    turns one long vector into storage space of lots of identically-sized
    smaller ones.

    Slice operations on the adapter cause new list objects (shallow copies) to 
    be created.

    If A, B, and C are different vectors stored in this adapter, then their
    ordering in the adaptee will be C{A0A1A2B0B1B2}, i.e. each vector is contiguous.
    We refer to this as 'vector-major' order.
    """

    __slots__ = ["adaptee", "unit", "length"]

    def __init__(self, adaptee, unit, length=None):
        self.adaptee = adaptee
        self.unit = unit
        self.length = length

        technical_len, remainder = divmod(len(self.adaptee), self.unit)
        assert remainder == 0

        if self.length is not None:
            assert self.length <= technical_len

    def __len__(self):
        if self.length is not None:
            return self.length
        else:
            technical_len, remainder = divmod(len(self.adaptee), self.unit)
            assert remainder == 0
            return technical_len

    def __iter__(parent):
        class FSSAIterator:
            def __init__(self):
                self.idx = 0
    
            def next(self):
                if self.idx >= len(parent):
                    raise StopIteration
                result = parent[self.idx]
                self.idx += 1
                return result
    
        return FSSAIterator()

    def __getitem__(self, idx):
        if isinstance(idx, int):
            if idx >= len(self):
                raise IndexError, idx
            return self.adaptee[self.unit*idx:self.unit*(idx+1)]
        elif isinstance(idx, slice):
            range_args= idx.indices(self.__len__())
            return [self.adaptee[self.unit*i:self.unit*(i+1)]
                    for i in xrange(*range_args)]
        else:
            raise TypeError, "invalid index type"

    def __setitem__(self, idx, value):
        if isinstance(idx, int):
            self.adaptee[self.unit*idx:self.unit*(idx+1)] = value
        elif isinstance(idx, slice):
            range_args= idx.indices(self.__len__())
            for i, subval in zip(xrange(*range_args), values):
                self.adaptee[self.unit*i:self.unit*(i+1)] = subval
        else:
            raise TypeError, "invalid index type"

    def get_alist_of_components(self):
        """Return the adaptee's data as an ArithmeticList of
        each vectors for each component.
        """
        from pytools.arithmetic_container import ArithmeticList
        return ArithmeticList(
                self.adaptee[i:len(self)*self.unit:self.unit] for i in range(self.unit))

    def get_component_major_vector(self):
        """Return the adaptee's data in component-major order.
        
        This gives a vector of order C{A0B0C0A1B1C1...}
        """
        return num.hstack(self.get_alist_of_components())





# index map tools -------------------------------------------------------------
def apply_index_map(imap, vector):
    from hedge._internal import VectorTarget, perform_index_map

    ls = log_shape(vector)
    if ls == ():
        result = numpy.zeros(shape=(imap.to_length,), dtype=float)
        perform_index_map(imap, VectorTarget(vector, result))
    else:
        result = numpy.zeros(shape=ls+(imap.to_length,), dtype=float)
        from pytools import indices_in_shape
        for i in indices_in_shape(ls):
            perform_index_map(imap, VectorTarget(vector[i], result[i]))
    return result




def apply_inverse_index_map(imap, vector):
    from hedge._internal import VectorTarget, perform_inverse_index_map

    ls = log_shape(vector)
    if ls == ():
        result = numpy.zeros(shape=(imap.from_length,), dtype=float)
        perform_inverse_index_map(imap, VectorTarget(vector, result))
    else:
        result = numpy.zeros(shape=ls+(imap.from_length,), dtype=float)
        from pytools import indices_in_shape
        for i in indices_in_shape(ls):
            perform_inverse_index_map(imap, VectorTarget(vector[i], result[i]))
    return result




# mesh reorderings ------------------------------------------------------------
def cuthill_mckee(graph):
    """Return a Cuthill-McKee ordering for the given graph.

    See (for example)
    Y. Saad, Iterative Methods for Sparse Linear System,
    2nd edition, p. 76.

    `graph' is given as an adjacency mapping, i.e. each node is
    mapped to a list of its neighbors.
    """
    from pytools import argmin

    # this list is called "old_numbers" because it maps a 
    # "new number to its "old number"
    old_numbers = []
    visited_nodes = set()
    levelset = []

    all_nodes = set(graph.keys())

    def levelset_cmp(node_a, node_b):
        return cmp(len(graph[node_a]), len(graph[node_b]))

    while len(old_numbers) < len(graph):
        if not levelset:
            unvisited = list(set(graph.keys()) - visited_nodes)

            if not unvisited:
                break

            start_node = unvisited[
                    argmin(len(graph[node]) for node in unvisited)]
            visited_nodes.add(start_node)
            old_numbers.append(start_node)
            levelset = [start_node]

        next_levelset = set()
        levelset.sort(levelset_cmp)

        for node in levelset:
            for neighbor in graph[node]:
                if neighbor in visited_nodes:
                    continue

                visited_nodes.add(neighbor)
                next_levelset.add(neighbor)
                old_numbers.append(neighbor)

        levelset = list(next_levelset)

    return old_numbers




def reverse_lookup_table(lut):
    result = [None] * len(lut)
    for key, value in enumerate(lut):
        result[value] = key
    return result




# block matrix ----------------------------------------------------------------
class BlockMatrix(object):
    """A block matrix is the sum of different smaller
    matrices positioned within one big matrix.
    """

    def __init__(self, chunks):
        """Return a new block matrix made up of components (`chunks')
        given as triples (i,j,smaller_matrix), where the top left (0,0)
        corner of the smaller_matrix is taken to be at position (i,j).

        smaller_matrix may be anything that can be left-multiplied to
        a Pylinear vector, including BlockMatrix instances.
        """
        self.chunks = []
        for i, j, chunk in chunks:
            if isinstance(chunk, BlockMatrix):
                self.chunks.extend(
                        (i+subi, j+subj, subchunk)
                        for subi, subj, subchunk in chunk.chunks)
            else:
                self.chunks.append((i, j, chunk))

    @property
    def T(self):
        return BlockMatrix(
                (j, i, chunk.T) for i, j, chunk in self.chunks)

    @property
    def H(self):
        return BlockMatrix(
                (j, i, chunk.H) for i, j, chunk in self.chunks)

    @property
    def shape(self):
        return (
                max(i+chunk.shape[0] for i, j, chunk in self.chunks),
                max(j+chunk.shape[0] for i, j, chunk in self.chunks)
                )

    def __add__(self, other):
        if isinstance(other, BlockMatrix):
            return BlockMatrix(self.chunks + other.chunks)
        else:
            return NotImplemented

    def __neg__(self):
        return BlockMatrix((i,j,-chunk) for i, j, chunk in self.chunks)

    def __sub__(self, other):
        if isinstance(other, BlockMatrix):
            return BlockMatrix(self.chunks + (-other).chunks)
        else:
            return NotImplemented

    def __mul__(self, other):
        if num.Vector.is_a(other):
            h, w = self.shape
            assert len(other) == w
            result = num.zeros((h,))

            for i, j, chunk in self.chunks:
                ch, cw = chunk.shape
                result[i:i+ch] += chunk * other[j:j+cw]

            return result
        elif isinstance(other, (float, complex, int)):
            return BlockMatrix(
                    (i,j,other*chunk) 
                    for i, j, chunk in self.chunks)
        else:
            return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (float, complex, int)):
            return BlockMatrix(
                    (i,j,other*chunk) 
                    for i, j, chunk in self.chunks)
        else:
            return NotImplemented

    def add_to_build_matrix(self, bmat):
        for i, j, chunk in self.chunks:
            bmat.add_block(i, j, chunk)




# parallel cg -----------------------------------------------------------------
try:
    import pyublasext
    OperatorBase = pyublasext.Operator(float)
except ImportError:
    class OperatorBase(object):
        @property
        def dtype(self):
            return float

        @property
        def shape(self):
            return (self.size1(), self.size2())

        def __call__(self, op2):
            temp = numpy.zeros((self.shape[0],), op2.dtype)
            self.apply(op2, temp)
            return temp





class CGStateContainer:
    def __init__(self, pcon, operator, precon=None):
        import pyublasext
        if precon is None:
            precon = pyublasext.IdentityOperator.make(operator.dtype, operator.size1())

        self.pcon = pcon
        self.operator = operator
        self.precon = precon

        if len(pcon.ranks) == 1:
            def inner(a, b):
                return numpy.dot(a, b.conj())
        else:
            from boost.mpi import all_reduce
            from operator import add

            def inner(a, b):
                local = numpy.dot(a, b.conj())
                return all_reduce(pcon.communicator, local, add)

        self.inner = inner

    def reset(self, rhs, x=None):
        self.rhs = rhs

        if x is None:
            x = numpy.zeros((self.operator.size1(),))
        self.x = x

        self.residual = rhs - self.operator(x)

        self.d = self.precon(self.residual)

        self.delta = self.inner(self.residual, self.d)
        return self.delta

    def one_iteration(self, compute_real_residual=False):
        # typed up from J.R. Shewchuk, 
        # An Introduction to the Conjugate Gradient Method
        # Without the Agonizing Pain, Edition 1 1/4 [8/1994]
        # Appendix B3

        q = self.operator(self.d)
        alpha = self.delta / self.inner(self.d, q)

        self.x += alpha * self.d

        if compute_real_residual:
            self.residual = self.rhs - self.operator(self.x)
        else:
            self.residual -= alpha*q

        s = self.precon(self.residual)
        delta_old = self.delta
        self.delta = self.inner(self.residual, s)

        beta = self.delta / delta_old;
        self.d = s + beta * self.d;

        return self.delta

    def run(self, max_iterations=None, tol=1e-7, debug_callback=None, debug=0):
        if max_iterations is None:
            max_iterations = 10 * self.operator.size1()

        if la.norm(self.rhs) == 0:
            return self.rhs

        iterations = 0
        delta_0 = delta = self.delta
        while iterations < max_iterations:
            if debug_callback is not None:
                debug_callback(self.x, self.residual, self.d)

            compute_real_residual = \
                    iterations % 50 == 0 or \
                    abs(delta) < tol*tol * abs(delta_0)
            delta = self.one_iteration(
                    compute_real_residual=compute_real_residual)

            if compute_real_residual and abs(delta) < tol*tol * abs(delta_0):
                return self.x

            if debug and iterations % debug == 0 and self.pcon.is_head_rank:
                print "debug: delta=%g" % delta
            iterations += 1

        raise RuntimeError("cg failed to converge")
            



def parallel_cg(pcon, operator, b, precon=None, x=None, tol=1e-7, max_iterations=None, 
        debug=False, debug_callback=None):
    if x is None:
        x = numpy.zeros((operator.size1(),))

    if len(pcon.ranks) == 1 and debug_callback is None:
        # use canned single-processor cg if possible
        import pyublasext
        a_inv = pyublasext.CGOperator.make(operator, max_it=max_iterations, 
                tolerance=tol, precon_op=precon)
        if debug:
            a_inv.debug_level = 1
        a_inv.apply(b, x)
        return x

    cg = CGStateContainer(pcon, operator, precon)
    cg.reset(b, x)
    return cg.run(max_iterations, tol, debug_callback, debug)



"""1D quadrature for Jacobi polynomials. Grundmann-Moeller cubature on the simplex."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import numpy




def legendre_gauss_quadrature(N):
    return jacobi_quadrature(0, 0, N)




def jacobi_gauss_lobatto_points(alpha, beta, N):
    """Compute the M{N}th order Gauss-Lobatto quadrature
    points, x, associated with the Jacobi polynomial,
    of type (alpha,beta) > -1 ( <> -0.5).
    """

    x = numpy.zeros((N+1,))
    x[0] = -1
    x[-1] = 1

    if N == 1:
        return x

    x[1:-1] = numpy.array(
            JacobiGaussQuadrature(alpha+1, beta+1, N-2).points
            ).real
    return x




def legendre_gauss_lobatto_points(N):
    """Compute the M{N}th order Gauss-Lobatto quadrature
    points, x, associated with the Legendre polynomials.
    """
    return jacobi_gauss_lobatto_points(0, 0, N)




class Quadrature(object):
    """An abstract quadrature rule."""
    def __init__(self, points, weights):
        self.weights = weights
        self.points = points
        self.data = zip(points, weights)

    def __call__(self, f):
        """Integrate the callable f with respect to the given quadrature rule.
        """
        return sum(w*f(x) for x, w in self.data)



class JacobiGaussQuadrature(Quadrature):
    """An M{N}th order Gauss quadrature associated with the Jacobi
    polynomials of type M{(alpha,beta) > -1} 
    
    C{alpha} and C{beta} may not be -0.5.
    """
    def __init__(self, alpha, beta, N):
        x, w = self.compute_weights_and_nodes(N, alpha, beta)
        Quadrature.__init__(self, x, w)

    @staticmethod
    def compute_weights_and_nodes(N, alpha, beta):
        """Return (nodes, weights) for an n-th order Gauss quadrature
        with the Jacobi polynomials of type (alpha, beta).
        """
        # follows 
        # Gene H. Golub, John H. Welsch, Calculation of Gauss Quadrature Rules, 
        # Mathematics of Computation, Vol. 23, No. 106 (Apr., 1969), pp. 221-230
        # doi:10.2307/2004418

        # see also doc/hedge-notes.tm for correspondence with the Jacobi
        # recursion from Hesthaven/Warburton's book

        from math import sqrt

        apb = alpha+beta

        # see Appendix A of Hesthaven/Warburton for these formulas
        def a(n):
            return (
                    2/(2*n+apb)
                    *
                    sqrt(
                        (n*(n+apb)*(n+alpha)*(n+beta))
                        /
                        ((2*n+apb-1)*(2*n+apb+1))
                        )
                    )

        def b(n):
            if n == 0:
                return (
                        -(alpha-beta)
                        /
                        (apb+2)
                        )
            else:
                return (
                        -(alpha**2-beta**2)
                        /
                        ((2*n+apb)*(2*n+apb+2))
                        )

        T = numpy.zeros((N+1, N+1))

        for n in range(N+1):
            T[n,n] = b(n)
            if n > 0:
                T[n,n-1] = current_a
            if n < N:
                next_a = a(n+1)
                T[n,n+1] = next_a
                current_a = next_a

        assert numpy.linalg.norm(T-T.T) < 1e-12
        eigval, eigvec = numpy.linalg.eigh(T)

        from numpy import dot, diag
        assert numpy.linalg.norm(dot(T, eigvec) -  dot(eigvec, diag(eigval))) < 1e-12

        from hedge.polynomial import JacobiFunction
        p0 = JacobiFunction(alpha, beta, 0)
        nodes = eigval
        weights = [eigvec[0,i]**2 / p0(nodes[i])**2 for i in range(N+1)]

        return nodes, weights


            

class LegendreGaussQuadrature(JacobiGaussQuadrature):
    """An M{N}th order Gauss quadrature associated with the Legendre polynomials.
    """
    def __init__(self, N):
        JacobiGaussQuadrature.__init__(self, 0, 0, N)




class TransformedQuadrature(Quadrature):
    """A quadrature rule on an arbitrary interval M{(a,b)}. """

    def __init__(self, quad, left, right):
        """Transform a given quadrature rule `quad' onto an arbitrary
        interval (left, right).
        """
        self.left = left
        self.right = right

        length = right-left
        assert length > 0
        half_length = length / 2
        Quadrature.__init__(self,
                [left + (p+1)/2*length for p in quad.points],
                [w*half_length for w in quad.weights])




def _extended_euclidean(q, r):
    """Return a tuple (p, a, b) such that p = aq + br, 
    where p is the greatest common divisor.
    """

    # see [Davenport], Appendix, p. 214

    if abs(q) < abs(r):
        p, a, b = _extended_euclidean(r, q)
        return p, b, a
  
    Q = 1, 0
    R = 0, 1
  
    while r:
        quot, t = divmod(q, r)
        T = Q[0] - quot*R[0], Q[1] - quot*R[1]
        q, r = r, t
        Q, R = R, T
  
    return q, Q[0], Q[1]




def _gcd(q, r):
    return _extended_euclidean(q, r)[0]




def _simplify_fraction((a, b)):
    gcd = _gcd(a,b)
    return (a//gcd, b//gcd)




class SimplexCubature(object):
    """Cubature on an M{n}-simplex.

    cf.
    A. Grundmann and H.M. Moeller,  
    Invariant integration formulas for the n-simplex by combinatorial methods, 
    SIAM J. Numer. Anal.  15 (1978), 282--290.

    This cubature rule has both negative and positive weights.
    """
    def __init__(self, order, dimension):
        s = order
        n = dimension
        d = 2*s+1

        from pytools import \
                generate_decreasing_nonnegative_tuples_summing_to, \
                generate_unique_permutations, \
                factorial, \
                wandering_element

        points_to_weights = {}

        for i in xrange(s+1):
            weight = (-1)**i * 2**(-2*s) \
                    * (d + n-2*i)**d \
                    / factorial(i) \
                    / factorial(d+n-i)

            for t in generate_decreasing_nonnegative_tuples_summing_to(s-i, n+1):
                for beta in generate_unique_permutations(t):
                    denominator = d+n-2*i
                    point = tuple(
                            _simplify_fraction((2*beta_i+1, denominator))
                            for beta_i in beta)

                    points_to_weights[point] = points_to_weights.get(point, 0) + weight

        from operator import add

        vertices = [-1 * numpy.ones((n,))] \
                + [numpy.array(x) for x in wandering_element(n, landscape=-1, wanderer=1)]

        self.pos_points = []
        self.pos_weights = []
        self.neg_points = []
        self.neg_weights = []

        dim_factor = 2**n
        for p, w in points_to_weights.iteritems():
            real_p = reduce(add, (a/b*v for (a,b),v in zip(p, vertices)))
            if w > 0:
                self.pos_points.append(real_p)
                self.pos_weights.append(dim_factor*w)
            else:
                self.neg_points.append(real_p)
                self.neg_weights.append(dim_factor*w)

        self.points = self.pos_points + self.neg_points
        self.weights = self.pos_weights + self.neg_weights

        self.pos_info = zip(self.pos_points, self.pos_weights)
        self.neg_info = zip(self.neg_points, self.neg_weights)


    def __call__(self, f):
        return sum(f(x)*w for x, w in self.pos_info) \
                + sum(f(x)*w for x, w in self.neg_info)




if __name__ == "__main__":
    cub = SimplexCubature(2, 3)
    print cub.weights
    for p in cub.points:
        print p



import scipy.special

maxnu = 10
n_zeros = 20

zeros = []
for n in range(0,maxnu+1):
    zeros.append(list(scipy.special.jn_zeros(n, n_zeros)))

outf = open("bessel_zeros.py", "w").write("bessel_zeros = %s" % zeros)



"""Local function space representation."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import numpy
import numpy.linalg as la
import pyublas
from hedge.tools import AffineMap
import hedge._internal
from math import sqrt, sin, cos, exp, pi
from pytools import memoize
import hedge.mesh




__all__ = ["TriangularElement", "TetrahedralElement"]




class WarpFactorCalculator:
    """Calculator for Warburton's warp factor.

    See T. Warburton,
    "An explicit construction of interpolation nodes on the simplex"
    Journal of Engineering Mathematics Vol 56, No 3, p. 247-262, 2006
    """

    def __init__(self, N):
        from hedge.quadrature import legendre_gauss_lobatto_points
        from hedge.interpolation import newton_interpolation_function

        # Find lgl and equidistant interpolation points
        r_lgl = legendre_gauss_lobatto_points(N)
        r_eq  = numpy.linspace(-1,1,N+1)

        self.int_f = newton_interpolation_function(r_eq, r_lgl - r_eq)

    def __call__(self, x):
        if abs(x) > 1-1e-10:
            return 0
        else:
            return self.int_f(x)/(1-x**2)




class FaceVertexMismatch(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)




class TriangleWarper:
    def __init__(self, alpha, order):
        self.alpha = alpha
        self.warp = WarpFactorCalculator(order)

        cls = TriangularElement

        from pytools import wandering_element
        from tools import normalize

        vertices = [cls.barycentric_to_equilateral(bary)
                for bary in wandering_element(cls.dimensions+1)]
        all_vertex_indices = range(cls.dimensions+1)
        face_vertex_indices = cls.geometry \
                .face_vertices(all_vertex_indices)
        faces_vertices = cls.geometry.face_vertices(vertices)

        edgedirs = [normalize(v2-v1) for v1, v2 in faces_vertices]
        opp_vertex_indices = [
            (set(all_vertex_indices) - set(fvi)).__iter__().next()
            for fvi in face_vertex_indices]

        self.loop_info = zip(
                face_vertex_indices, 
                edgedirs, 
                opp_vertex_indices)

    def __call__(self, bp):
        shifts = []

        from operator import add, mul

        for fvi, edgedir, opp_vertex_index in self.loop_info:
            blend = 4*reduce(mul, (bp[i] for i in fvi))
            warp_amount = blend*self.warp(bp[fvi[1]]-bp[fvi[0]]) \
                    * (1 + (self.alpha*bp[opp_vertex_index])**2)
            shifts.append(warp_amount*edgedir)

        return reduce(add, shifts)




TriangleBasisFunction = hedge._internal.TriangleBasisFunction
GradTriangleBasisFunction = hedge._internal.GradTriangleBasisFunction
TetrahedronBasisFunction = hedge._internal.TetrahedronBasisFunction
GradTetrahedronBasisFunction = hedge._internal.GradTetrahedronBasisFunction




class Element(object):
    def basis_functions(self):
        """Get a sequence of functions that form a basis of the approximation space."""
        raise NotImplementedError

    def grad_basis_functions(self):
        """Get the gradient functions of the basis_functions(), in the same order."""
        raise NotImplementedError

    # matrices ----------------------------------------------------------------
    @memoize
    def vandermonde(self):
        from hedge.polynomial import generic_vandermonde

        return generic_vandermonde(
                list(self.unit_nodes()),
                list(self.basis_functions()))

    @memoize
    def inverse_mass_matrix(self):
        """Return the inverse of the mass matrix of the unit element 
        with respect to the nodal coefficients. Divide by the Jacobian 
        to obtain the global mass matrix.
        """

        # see doc/hedge-notes.tm
        v = self.vandermonde()
        return numpy.dot(v, v.T)

    @memoize
    def mass_matrix(self):
        """Return the mass matrix of the unit element with respect 
        to the nodal coefficients. Multiply by the Jacobian to obtain
        the global mass matrix.
        """

        return numpy.asarray(la.inv(self.inverse_mass_matrix()), order="C")

    @memoize
    def grad_vandermonde(self):
        """Compute the Vandermonde matrices of the grad_basis_functions().
        Return a list of these matrices."""

        from hedge.polynomial import generic_multi_vandermonde

        return generic_multi_vandermonde(
                list(self.unit_nodes()),
                list(self.grad_basis_functions()))

    @memoize
    def differentiation_matrices(self):
        """Return matrices that map the nodal values of a function
        to the nodal values of its derivative in each of the unit
        coordinate directions.
        """

        from hedge.tools import leftsolve
        # see doc/hedge-notes.tm
        v = self.vandermonde()
        return [leftsolve(v, vdiff) for vdiff in self.grad_vandermonde()]




class SimplicialElement(Element):
    # numbering ---------------------------------------------------------------
    def node_count(self):
        """Return the number of interpolation nodes in this element."""
        d = self.dimensions
        o = self.order
        from operator import mul
        from pytools import factorial
        return int(reduce(mul, (o+1+i for i in range(d)))/factorial(d))

    @memoize
    def vertex_indices(self):
        """Return the list of the vertices' node indices."""
        from pytools import wandering_element

        result = []

        node_tup_to_idx = dict(
                (ituple, idx) 
                for idx, ituple in enumerate(self.node_tuples()))

        vertex_tuples = [self.dimensions * (0,)] \
                + list(wandering_element(self.dimensions, wanderer=self.order))

        return [node_tup_to_idx[vt] for vt in vertex_tuples]

    @memoize
    def face_indices(self):
        """Return a list of face index lists. Each face index list contains
        the local node numbers of the nodes on that face.
        """

        node_tup_to_idx = dict(
                (ituple, idx) 
                for idx, ituple in enumerate(self.node_tuples()))

        from pytools import generate_nonnegative_integer_tuples_summing_to_at_most

        enum_order_nodes_gen = generate_nonnegative_integer_tuples_summing_to_at_most(
                    self.order, self.dimensions)

        faces = [[] for i in range(self.dimensions+1)]

        for node_tup in enum_order_nodes_gen:
            for face_idx in self.faces_for_node_tuple(node_tup):
                faces[face_idx].append(node_tup_to_idx[node_tup])

        return [tuple(fi) for fi in faces]

    # node wrangling ----------------------------------------------------------
    def equidistant_barycentric_nodes(self):
        """Generate equidistant nodes in barycentric coordinates."""
        for indices in self.node_tuples():
            divided = tuple(i/self.order for i in indices)
            yield (1-sum(divided),) + divided

    def equidistant_equilateral_nodes(self):
        """Generate equidistant nodes in equilateral coordinates."""

        for bary in self.equidistant_barycentric_nodes():
            yield self.barycentric_to_equilateral(bary)

    def equidistant_unit_nodes(self):
        """Generate equidistant nodes in unit coordinates."""

        for bary in self.equidistant_barycentric_nodes():
            yield self.equilateral_to_unit(self.barycentric_to_equilateral(bary))

    @memoize
    def unit_nodes(self):
        """Generate the warped nodes in unit coordinates (r,s,...)."""
        return [self.equilateral_to_unit(node)
                for node in self.equilateral_nodes()]

    # basis functions ---------------------------------------------------------
    def generate_mode_identifiers(self):
        """Generate a hashable objects identifying each basis function, in order.

        The output from this function is required to be in the same order
        as that of L{basis_functions} and L{grad_basis_functions}, and thereby
        also from L{vandermonde}.
        """
        from pytools import generate_nonnegative_integer_tuples_summing_to_at_most

        return generate_nonnegative_integer_tuples_summing_to_at_most(
                self.order, self.dimensions)

    # time step scaling -------------------------------------------------------
    def dt_non_geometric_factor(self):
        unodes = self.unit_nodes()
        vertex_indices = self.vertex_indices()

        return 2/3*\
                min(min(min(
                    la.norm(unodes[face_node_index]-unodes[vertex_index])
                    for vertex_index in vertex_indices
                    if vertex_index != face_node_index)
                    for face_node_index in face_indices)
                    for face_indices in self.face_indices())





class IntervalElementBase(SimplicialElement):
    dimensions = 1
    has_local_jacobians = False
    geometry = hedge.mesh.Interval

    # numbering ---------------------------------------------------------------
    @memoize
    def node_tuples(self):
        """Generate tuples enumerating the node indices present
        in this element. Each tuple has a length equal to the dimension
        of the element. The tuples constituents are non-negative integers
        whose sum is less than or equal to the order of the element.
        
        The order in which these nodes are generated dictates the local 
        node numbering.
        """
        return [(i,) for i in range(self.order+1)]

    def faces_for_node_tuple(self, node_idx):
        """Return the list of face indices of faces on which the node 
        represented by C{node_idx} lies.
        """

        if node_idx == (0,):
            return [0]
        elif node_idx == (self.order,):
            return [1]
        else:
            return []

    # node wrangling ----------------------------------------------------------
    @memoize
    def get_submesh_indices(self):
        """Return a list of tuples of indices into the node list that
        generate a tesselation of the reference element."""

        return [(i,i+1) for i in range(self.order)]

    # face operations ---------------------------------------------------------
    @memoize
    def face_mass_matrix(self):
        return numpy.array([[1]], dtype=float)

    @staticmethod
    def get_face_index_shuffle_to_match(face_1_vertices, face_2_vertices):
        if set(face_1_vertices) != set(face_2_vertices):
            raise FaceVertexMismatch("face vertices do not match")

        class IntervalFaceIndexShuffle:
            def __hash__(self):
                return 0x3472477

            def __eq__(self, other):
                return True

            def __call__(self, indices):
                return indices

        return IntervalFaceIndexShuffle()





class IntervalElement(IntervalElementBase):
    """An arbitrary-order polynomial finite interval element.

    Coordinate systems used:
    ========================

    unit coordinates (r)::

    ---[--------0--------]--->
       -1                1
    """

    def __init__(self, order):
        self.order = order

    # node wrangling ----------------------------------------------------------
    def nodes(self):
        """Generate warped nodes in unit coordinates (r,)."""

        from hedge.quadrature import legendre_gauss_lobatto_points
        return [numpy.array([x]) for x in legendre_gauss_lobatto_points(self.order)]

    equilateral_nodes = nodes
    unit_nodes = nodes

    # basis functions ---------------------------------------------------------
    @memoize
    def basis_functions(self):
        """Get a sequence of functions that form a basis of the approximation space.

        The approximation space is spanned by the polynomials:::

          r**i for i <= N
        """
        from hedge.polynomial import LegendreFunction

        class VectorLF:
            def __init__(self, n):
                self.lf = LegendreFunction(n)

            def __call__(self, x):
                return self.lf(x[0])

        return [VectorLF(idx[0]) for idx in self.generate_mode_identifiers()]

    def grad_basis_functions(self):
        """Get the gradient functions of the basis_functions(), in the same order."""
        from hedge.polynomial import DiffLegendreFunction

        class DiffVectorLF:
            def __init__(self, n):
                self.dlf = DiffLegendreFunction(n)

            def __call__(self, x):
                return numpy.array([self.dlf(x[0])])

        return [DiffVectorLF(idx[0]) 
            for idx in self.generate_mode_identifiers()]

    # time step scaling -------------------------------------------------------
    def dt_non_geometric_factor(self):
        unodes = self.unit_nodes()
        return la.norm(unodes[0] - unodes[1])

    def dt_geometric_factor(self, vertices, el):
        return abs(el.map.jacobian)




class TriangularElement(SimplicialElement):
    """An arbitrary-order triangular finite element.

    Coordinate systems used:
    ========================

    unit coordinates (r,s)::

    C
    |\\
    | \\
    |  O
    |   \\
    |    \\
    A-----B

    Points in unit coordinates::

        O = (0,0)
        A = (-1,-1)
        B = (1,-1)
        C = (-1,1)

    equilateral coordinates (x,y)::

            C
           / \\
          /   \\
         /     \\
        /   O   \\
       /         \\
      A-----------B 

    Points in equilateral coordinates::

        O = (0,0)
        A = (-1,-1/sqrt(3))
        B = (1,-1/sqrt(3))
        C = (0,2/sqrt(3))

    When global vertices are passed in, they are mapped to the 
    reference vertices A, B, C in order.

    Faces are always ordered AB, BC, AC.
    """

    # In case you were wondering: the double backslashes in the docstring
    # are required because single backslashes only escape their subsequent
    # newlines, and thus end up not yielding a correct docstring.

    dimensions = 2
    has_local_jacobians = False
    geometry = hedge.mesh.Triangle

    def __init__(self, order):
        self.order = order

    # numbering ---------------------------------------------------------------
    @memoize
    def node_tuples(self):
        """Generate tuples enumerating the node indices present
        in this element. Each tuple has a length equal to the dimension
        of the element. The tuples constituents are non-negative integers
        whose sum is less than or equal to the order of the element.
        
        The order in which these nodes are generated dictates the local 
        node numbering.
        """
        from pytools import generate_nonnegative_integer_tuples_summing_to_at_most
        node_tups = list(generate_nonnegative_integer_tuples_summing_to_at_most(
                self.order, self.dimensions))

        faces_to_nodes = {}
        for node_tup in node_tups:
            faces_to_nodes.setdefault(
                    frozenset(self.faces_for_node_tuple(node_tup)),
                    []).append(node_tup)

        result = []
        def add_face_nodes(faces):
            result.extend(faces_to_nodes.get(frozenset(faces), []))

        add_face_nodes([])
        add_face_nodes([0])
        add_face_nodes([0,1])
        add_face_nodes([1])
        add_face_nodes([1,2])
        add_face_nodes([2])
        add_face_nodes([0,2])

        assert set(result) == set(node_tups)
        assert len(result) == len(node_tups)

        return result

    def faces_for_node_tuple(self, node_tuple):
        """Return the list of face indices of faces on which the node 
        represented by C{node_tuple} lies.
        """
        m, n = node_tuple

        result = []
        if n == 0:
            result.append(0)
        if n+m == self.order:
            result.append(1)
        if m == 0:
            result.append(2)
        return result

    # node wrangling ----------------------------------------------------------
    @staticmethod
    def barycentric_to_equilateral((lambda1, lambda2, lambda3)):
        """Return the equilateral (x,y) coordinate corresponding
        to the barycentric coordinates (lambda1..lambdaN)."""

        # reflects vertices in equilateral coordinates
        return numpy.array([
            (-lambda1  +lambda2            ),
            (-lambda1  -lambda2  +2*lambda3)/sqrt(3.0)])

    # see doc/hedge-notes.tm
    equilateral_to_unit = AffineMap(
            numpy.array([[1,-1/sqrt(3)], [0,2/sqrt(3)]]),
                numpy.array([-1/3,-1/3]))

    def equilateral_nodes(self):
        """Generate warped nodes in equilateral coordinates (x,y)."""

        # port of Hesthaven/Warburton's Nodes2D routine
        # note that the order of the barycentric coordinates is changed
        # match the order of the equilateral vertices 
        
        # Not much is left of the original routine--it was very redundant.
        # The test suite still contains the original code and verifies this
        # one against it.

        # Set optimized parameter alpha, depending on order N
        alpha_opt = [0.0000, 0.0000, 1.4152, 0.1001, 0.2751, 0.9800, 1.0999,
                1.2832, 1.3648, 1.4773, 1.4959, 1.5743, 1.5770, 1.6223, 1.6258]
                  
        try:
            alpha = alpha_opt[self.order-1]
        except IndexError:
            alpha = 5/3

        warp = TriangleWarper(alpha, self.order)

        for bp in self.equidistant_barycentric_nodes():
            yield self.barycentric_to_equilateral(bp) + warp(bp)

    @memoize
    def get_submesh_indices(self):
        """Return a list of tuples of indices into the node list that
        generate a tesselation of the reference element."""

        node_dict = dict(
                (ituple, idx) 
                for idx, ituple in enumerate(self.node_tuples()))

        result = []
        for i, j in self.node_tuples():
            if i+j < self.order:
                result.append(
                        (node_dict[i,j], node_dict[i+1,j], node_dict[i,j+1]))
            if i+j < self.order-1:
                result.append(
                    (node_dict[i+1,j+1], node_dict[i,j+1], node_dict[i+1,j]))
        return result

    # basis functions ---------------------------------------------------------
    @memoize
    def basis_functions(self):
        """Get a sequence of functions that form a basis of the approximation space.

        The approximation space is spanned by the polynomials:::

          r**i * s**j for i+j <= N
        """
        return [TriangleBasisFunction(*idx) for idx in 
                self.generate_mode_identifiers()]

    def grad_basis_functions(self):
        """Get the gradient functions of the basis_functions(), in the same order."""
        return [GradTriangleBasisFunction(*idx) for idx in 
                self.generate_mode_identifiers()]

    # face operations ---------------------------------------------------------
    @memoize
    def face_mass_matrix(self):
        from hedge.polynomial import legendre_vandermonde
        unodes = self.unit_nodes()
        face_vandermonde = legendre_vandermonde(
                [unodes[i][0] for i in self.face_indices()[0]],
                self.order)

        return numpy.asarray(
                la.inv(
                    numpy.dot(face_vandermonde, face_vandermonde.T)),
                order="C")

    @staticmethod
    def get_face_index_shuffle_to_match(face_1_vertices, face_2_vertices):
        if set(face_1_vertices) != set(face_2_vertices):
            raise FaceVertexMismatch("face vertices do not match")

        class TriangleFaceIndexShuffle:
            def __init__(self, operations):
                self.operations = operations

            def __hash__(self):
                return hash(self.operations)

            def __eq__(self, other):
                return self.operations == other.operations

            def __call__(self, indices):
                for op in self.operations:
                    if op == "flip":
                        indices = indices[::-1]
                    else:
                        raise RuntimeError, "invalid operation"
                return indices


        if face_1_vertices != face_2_vertices:
            assert face_1_vertices[::-1] == face_2_vertices
            return TriangleFaceIndexShuffle(("flip",))
        else:
            return TriangleFaceIndexShuffle(())

    # time step scaling -------------------------------------------------------
    def dt_geometric_factor(self, vertices, el):
        area = abs(2*el.map.jacobian)
        semiperimeter = sum(la.norm(vertices[vi1]-vertices[vi2]) 
                for vi1, vi2 in [(0,1), (1,2), (2,0)])/2
        return area/semiperimeter




class TetrahedralElement(SimplicialElement):
    """An arbitrary-order tetrahedral finite element.

    Coordinate systems used:
    ========================

    unit coordinates (r,s,t)::

               ^ s
               |
               C
              /|\\
             / | \\
            /  |  \\
           /   |   \\
          /   O|    \\
         /   __A-----B---> r
        /_--^ ___--^^
       ,D--^^^
    t L 
      
    (squint, and it might start making sense...)

    Points in unit coordinates::

        O=( 0, 0, 0)
        A=(-1,-1,-1)
        B=(+1,-1,-1)
        C=(-1,+1,-1)
        D=(-1,-1,+1)

    Points in equilateral coordinates (x,y,z)::

        O = (0,0)
        A = (-1,-1/sqrt(3),-1/sqrt(6))
        B = ( 1,-1/sqrt(3),-1/sqrt(6))
        C = ( 0, 2/sqrt(3),-1/sqrt(6))
        D = ( 0,         0, 3/sqrt(6))

    When global vertices are passed in, they are mapped to the 
    reference vertices A, B, C, D in order.

    Faces are always ordered ABC, ABD, ACD, BCD.
    """

    # In case you were wondering: the double backslashes in the docstring
    # above are required because single backslashes only escape their subsequent
    # newlines, and thus end up not yielding a correct docstring.

    dimensions = 3
    has_local_jacobians = False
    geometry = hedge.mesh.Tetrahedron

    def __init__(self, order):
        self.order = order

    # numbering ---------------------------------------------------------------
    @memoize
    def node_tuples(self):
        """Generate tuples enumerating the node indices present
        in this element. Each tuple has a length equal to the dimension
        of the element. The tuples constituents are non-negative integers
        whose sum is less than or equal to the order of the element.
        
        The order in which these nodes are generated dictates the local 
        node numbering.
        """
        from pytools import generate_nonnegative_integer_tuples_summing_to_at_most
        node_tups = list(generate_nonnegative_integer_tuples_summing_to_at_most(
                self.order, self.dimensions))

        if False:
            # hand-tuned node order
            faces_to_nodes = {}
            for node_tup in node_tups:
                faces_to_nodes.setdefault(
                        frozenset(self.faces_for_node_tuple(node_tup)),
                        []).append(node_tup)

            result = []
            def add_face_nodes(faces):
                result.extend(faces_to_nodes.get(frozenset(faces), []))

            add_face_nodes([0,3])
            add_face_nodes([0])
            add_face_nodes([0,2])
            add_face_nodes([0,1])
            add_face_nodes([0,1,2])
            add_face_nodes([0,1,3])
            add_face_nodes([0,2,3])
            add_face_nodes([1])
            add_face_nodes([1,2])
            add_face_nodes([1,2,3])
            add_face_nodes([1,3])
            add_face_nodes([2])
            add_face_nodes([2,3])
            add_face_nodes([3])
            add_face_nodes([])

            assert set(result) == set(node_tups)
            assert len(result) == len(node_tups)

        if True:
            # average-sort heuristic node order
            from pytools import average

            def order_number_for_node_tuple(nt):
                faces = self.faces_for_node_tuple(nt)
                if not faces:
                    return -1
                elif len(faces) >= 3:
                    return 1000
                else:
                    return average(faces)

            def cmp_node_tuples(nt1, nt2):
                return cmp(
                        order_number_for_node_tuple(nt1), 
                        order_number_for_node_tuple(nt2))

            result = node_tups
            #result.sort(cmp_node_tuples)

        for i, nt in enumerate(result):
            fnt = self.faces_for_node_tuple(nt)
            #print i, nt, fnt

        return result

    def faces_for_node_tuple(self, node_tuple):
        """Return the list of face indices of faces on which the node 
        represented by C{node_tuple} lies.
        """
        m,n,o = node_tuple
        result = []

        if o == 0:
            result.append(0)
        if n == 0:
            result.append(1)
        if m == 0:
            result.append(2)
        if n+m+o == self.order:
            result.append(3)

        return result

    # node wrangling ----------------------------------------------------------
    @staticmethod
    def barycentric_to_equilateral((lambda1, lambda2, lambda3, lambda4)):
        """Return the equilateral (x,y) coordinate corresponding
        to the barycentric coordinates (lambda1..lambdaN)."""

        # reflects vertices in equilateral coordinates
        return numpy.array([
            (-lambda1  +lambda2                        ),
            (-lambda1  -lambda2  +2*lambda3            )/sqrt(3.0),
            (-lambda1  -lambda2  -  lambda3  +3*lambda4)/sqrt(6.0),
            ])

    # see doc/hedge-notes.tm
    equilateral_to_unit = AffineMap(
            numpy.array([
                [1,-1/sqrt(3),-1/sqrt(6)], 
                [0, 2/sqrt(3),-1/sqrt(6)],
                [0,         0,   sqrt(6)/2]
                ]),
                numpy.array([-1/2,-1/2,-1/2]))

    def equilateral_nodes(self):
        """Generate warped nodes in equilateral coordinates (x,y)."""

        # port of Hesthaven/Warburton's Nodes3D routine

        # Set optimized parameter alpha, depending on order N
        alpha_opt = [0,0,0,0.1002, 1.1332,1.5608,1.3413,1.2577,1.1603,
                1.10153,0.6080,0.4523,0.8856,0.8717,0.9655]

        try:
            alpha = alpha_opt[self.order-1]
        except IndexError:
            alpha = 1

        from pytools import wandering_element

        vertices = [self.barycentric_to_equilateral(bary)
                for bary in wandering_element(self.dimensions+1)]
        all_vertex_indices = range(self.dimensions+1)
        face_vertex_indices = self.geometry \
                .face_vertices(all_vertex_indices)
        faces_vertices = self.geometry \
                .face_vertices(vertices)

        bary_points = list(self.equidistant_barycentric_nodes())
        equi_points = [self.barycentric_to_equilateral(bp) 
                for bp in bary_points]

        from tools import normalize
        from operator import add, mul

        tri_warp = TriangleWarper(alpha, self.order)

        for fvi, (v1, v2, v3) in zip(face_vertex_indices, faces_vertices):
            # find directions spanning the face: "base" and "altitude"
            directions = [normalize(v2-v1), normalize((v3)-(v1+v2)/2)]

            # the two should be orthogonal
            assert abs(numpy.dot(directions[0],directions[1])) < 1e-16

            # find the vertex opposite to the current face
            opp_vertex_index = (set(all_vertex_indices) - set(fvi)).__iter__().next()

            shifted = []
            for bp, ep in zip(bary_points, equi_points):
                face_bp = [bp[i] for i in fvi]

                blend = reduce(mul, face_bp) * (1+alpha*bp[opp_vertex_index])**2

                for i in fvi:
                    denom = bp[i] + 0.5*bp[opp_vertex_index]
                    if abs(denom) > 1e-12:
                        blend /= denom
                    else:
                        blend = 0.5 # each edge gets shifted twice
                        break
                    
                shifted.append(ep + blend*reduce(add,
                    (tw*dir for tw, dir in zip(tri_warp(face_bp), directions))))

            equi_points = shifted

        return equi_points

    @memoize
    def get_submesh_indices(self):
        """Return a list of tuples of indices into the node list that
        generate a tesselation of the reference element."""

        node_dict = dict(
                (ituple, idx) 
                for idx, ituple in enumerate(self.node_tuples()))

        def add_tuples(a, b):
            return tuple(ac+bc for ac, bc in zip(a,b))

        def try_add_tet(d1, d2, d3, d4):
            try:
                result.append((
                    node_dict[add_tuples(current, d1)],
                    node_dict[add_tuples(current, d2)],
                    node_dict[add_tuples(current, d3)],
                    node_dict[add_tuples(current, d4)],
                    ))
            except KeyError, e:
                pass

        result = []
        for current in self.node_tuples():
            # this is a tesselation of a cube into six tets.
            # subtets that fall outside of the master tet are simply not added.

            # positively oriented
            try_add_tet((0,0,0), (1,0,0), (0,1,0), (0,0,1))
            try_add_tet((1,0,1), (1,0,0), (0,0,1), (0,1,0))
            try_add_tet((1,0,1), (0,1,1), (0,1,0), (0,0,1))

            try_add_tet((1,0,0), (0,1,0), (1,0,1), (1,1,0))
            try_add_tet((0,1,1), (0,1,0), (1,1,0), (1,0,1))
            try_add_tet((0,1,1), (1,1,1), (1,0,1), (1,1,0))

            # negatively oriented
            #try_add_tet((0,0,0), (1,0,0), (0,0,1), (0,1,0))
            #try_add_tet((1,0,1), (1,0,0), (0,1,0), (0,0,1))
            #try_add_tet((1,0,1), (0,1,1), (0,0,1), (0,1,0))

            #try_add_tet((1,0,0), (0,1,0), (1,1,0), (1,0,1))
            #try_add_tet((0,1,1), (0,1,0), (1,0,1), (1,1,0))
            #try_add_tet((0,1,1), (1,1,1), (1,1,0), (1,0,1))
        return result

    # basis functions ---------------------------------------------------------
    @memoize
    def basis_functions(self):
        """Get a sequence of functions that form a basis of the approximation space.

        The approximation space is spanned by the polynomials::

          r**i * s**j * t**k  for  i+j+k <= order
        """
        return [TetrahedronBasisFunction(*idx) for idx in 
                self.generate_mode_identifiers()]

    def grad_basis_functions(self):
        """Get the (r,s,...) gradient functions of the basis_functions(),
        in the same order.
        """
        return [GradTetrahedronBasisFunction(*idx) for idx in 
                self.generate_mode_identifiers()]

    # face operations ---------------------------------------------------------
    @memoize
    def face_mass_matrix(self):
        from hedge.polynomial import generic_vandermonde

        from pytools import generate_nonnegative_integer_tuples_summing_to_at_most

        basis = [TriangleBasisFunction(*node_tup) 
                for node_tup in 
                generate_nonnegative_integer_tuples_summing_to_at_most(
                    self.order, self.dimensions-1)]

        unodes = self.unit_nodes()
        face_indices = self.face_indices()

        face_vandermonde = generic_vandermonde(
                [unodes[i][:2] for i in face_indices[0]],
                basis)

        return numpy.asarray(
                la.inv(
                    numpy.dot(face_vandermonde, face_vandermonde.T)),
                order="C")

    def get_face_index_shuffle_to_match(self, face_1_vertices, face_2_vertices):
        (a,b,c) = face_1_vertices
        f2_tuple = face_2_vertices

        try:
            idx_map = self._shuffle_face_idx_map
        except AttributeError:
            idx_map = self._shuffle_face_idx_map = {}
            idx = 0
            for j in range(0, self.order+1):
                for i in range(0, self.order+1-j):
                    self._shuffle_face_idx_map[i,j] = idx
                    idx += 1

        order = self.order

        class TetrahedronFaceIndexShuffle:
            def __init__(self, operations):
                self.operations = operations

            def __hash__(self):
                return hash(self.operations)

            def __eq__(self, other):
                return self.operations == other.operations

            def __call__(self, indices):
                for op in self.operations:
                    if op == "flip":
                        indices = self.flip(indices)
                    elif op == "shift_left":
                        indices = self.shift_left(indices)
                    else:
                        raise RuntimeError, "invalid operation"
                return indices

            # flip and shift_left generate S_3
            def flip(self, indices):
                """Flip the indices along the unit hypotenuse."""
                result = []
                for j in range(0, order+1):
                    for i in range(0, order+1-j):
                        result.append(indices[idx_map[j,i]])
                return result

            def shift_left(self, indices):
                """Rotate all edges to the left."""
                result = len(indices)*[0]
                idx = 0
                for j in range(0, order+1):
                    for i in range(0, order+1-j):
                        result[idx_map[j, order-i-j]] = indices[idx]
                        idx += 1
                return result

        # yay, enumerate S_3 by hand
        if f2_tuple == (a,b,c):
            #return face_2_indices
            return TetrahedronFaceIndexShuffle(())
        elif f2_tuple == (a,c,b):
            #return flip(face_2_indices)
            return TetrahedronFaceIndexShuffle(("flip",))
        elif f2_tuple == (b,c,a):
            # (b,c,a) -sl-> (c,a,b) -sl-> (a,b,c)
            #return shift_left(shift_left(face_2_indices))
            return TetrahedronFaceIndexShuffle(("shift_left", "shift_left"))
        elif f2_tuple == (b,a,c):
            # (b,a,c) -sl-> (a,c,b) -fl-> (a,b,c)
            #return flip(shift_left(face_2_indices))
            return TetrahedronFaceIndexShuffle(("shift_left", "flip"))
        elif f2_tuple == (c,a,b):
            # (c,a,b) -sl-> (a,b,c)
            #return shift_left(face_2_indices)
            return TetrahedronFaceIndexShuffle(("shift_left",))
        elif f2_tuple == (c,b,a):
            # (c,b,a) -fl-> (c,a,b) -sl-> (a,b,c)
            #return shift_left(flip(face_2_indices))
            return TetrahedronFaceIndexShuffle(("flip", "shift_left"))
        else:
            raise FaceVertexMismatch("face vertices do not match")

    # time step scaling -------------------------------------------------------
    def dt_geometric_factor(self, vertices, el):
        return abs(el.map.jacobian)/max(abs(fj) for fj in el.face_jacobians)





#!/usr/bin/env python
# -*- coding: latin-1 -*-

# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




def get_config_schema():
    from aksetup_helper import ConfigSchema, Option, \
            IncludeDir, LibraryDir, Libraries, \
            Switch, StringListOption

    return ConfigSchema([
        IncludeDir("BOOST", []),
        LibraryDir("BOOST", []),
        Libraries("BOOST_PYTHON", ["boost_python-gcc42-mt"]),

        IncludeDir("NUMPY"),

        IncludeDir("BOOST_BINDINGS", []),

        Switch("HAVE_BLAS", False, "Whether to build with support for BLAS"),
        LibraryDir("BLAS", []),
        Libraries("BLAS", ["blas"]),

        Switch("HAVE_MPI", False, "Whether to build with support for MPI"),
        Option("MPICC", "mpicc",
            "Path to MPI C compiler"),
        Option("MPICXX", 
            help="Path to MPI C++ compiler (defaults to same as MPICC)"),
        Libraries("BOOST_MPI", ["boost_mpi-gcc42-mt"]),

        StringListOption("CXXFLAGS", [], 
            help="Any extra C++ compiler options to include"),
        StringListOption("LDFLAGS", [], 
            help="Any extra linker options to include"),
        ])




def main():
    import glob
    from aksetup_helper import hack_distutils, get_config, setup, Extension

    hack_distutils()
    conf = get_config()

    LIBRARY_DIRS = conf["BOOST_LIB_DIR"]
    LIBRARIES = conf["BOOST_PYTHON_LIBNAME"]

    EXTRA_DEFINES = {}
    EXTRA_INCLUDE_DIRS = []
    EXTRA_LIBRARY_DIRS = []
    EXTRA_LIBRARIES = []

    if conf["HAVE_MPI"]:
        EXTRA_DEFINES["USE_MPI"] = 1
        EXTRA_DEFINES["OMPI_SKIP_MPICXX"] = 1
        LIBRARIES.extend(conf["BOOST_MPI_LIBNAME"])

        from distutils import sysconfig
        cvars = sysconfig.get_config_vars()
        cvars["CC"] = conf["MPICC"]
        cvars["CXX"] = conf["MPICXX"]

    if conf["NUMPY_INC_DIR"] is None:
        try:
            import numpy
            from os.path import join
            conf["NUMPY_INC_DIR"] = [join(numpy.__path__[0], "core", "include")]
        except:
            pass

    INCLUDE_DIRS = [
            "src/cpp",
            ] \
            + conf["BOOST_BINDINGS_INC_DIR"] \
            + conf["BOOST_INC_DIR"] \
            + conf["NUMPY_INC_DIR"]

    conf["BLAS_INC_DIR"] = []
    conf["USE_BLAS"] = conf["HAVE_BLAS"]

    def handle_component(comp):
        if conf["USE_"+comp]:
            EXTRA_DEFINES["USE_"+comp] = 1
            EXTRA_INCLUDE_DIRS.extend(conf[comp+"_INC_DIR"])
            EXTRA_LIBRARY_DIRS.extend(conf[comp+"_LIB_DIR"])
            EXTRA_LIBRARIES.extend(conf[comp+"_LIBNAME"])

    handle_component("BLAS")

    setup(name="hedge",
            # metadata
            version="0.90",
            description="Hybrid Easy Discontinuous Galerkin Environment",
            long_description="""
            hedge is an unstructured, high-order, parallel
            Discontinuous Galerkin solver for partial differential
            equations.   
            
            Features:

            * Supports simplicial unstructured meshes in two and
              three dimensions (i.e. triangles and tetrahedra)
            * Approximates using orthogonal polynomials of any degree
              (and therefore to any order of accuracy) you specify at
              runtime
            * Solves PDEs in parallel using MPI 
            * Easy to use
            * Contains powerful parallel visualization features
            """,
            author=u"Andreas Kloeckner",
            author_email="inform@tiker.net",
            license = "GPLv3",
            url="http://mathema.tician.de/software/hedge",
            classifiers=[
              'Environment :: Console',
              'Development Status :: 4 - Beta',
              'Intended Audience :: Developers',
              'Intended Audience :: Other Audience',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: GNU General Public License (GPL)',
              'Natural Language :: English',
              'Programming Language :: C++',
              'Programming Language :: Python',
              'Topic :: Scientific/Engineering',
              'Topic :: Scientific/Engineering :: Mathematics',
              'Topic :: Scientific/Engineering :: Physics',
              'Topic :: Scientific/Engineering :: Visualization',
              ],

            # dependencies
            setup_requires=[
                "PyUblas>=0.92",
                ],
            install_requires=[
                "pytools>=2",
                "pymbolic>=0.90",
                "meshpy>=0.91",
                ],
            extras_require = {
                "elliptic": ["PyUblasExt>=0.92"],
                "silo": ["pylo"],
                "parallel": ["PyMetis>=0.91"],
                },

            # build info
            packages=["hedge"],
            zip_safe=False,

            package_dir={"hedge": "src/python"},
            ext_package="hedge",

            ext_modules=[
                Extension("_internal", 
                    ["src/wrapper/wrap_main.cpp", 
                        "src/wrapper/wrap_base.cpp", 
                        "src/wrapper/wrap_special_function.cpp", 
                        "src/wrapper/wrap_flux.cpp", 
                        "src/wrapper/wrap_op_target.cpp", 
                        "src/wrapper/wrap_volume_operators.cpp", 
                        "src/wrapper/wrap_index_map.cpp", 
                        "src/wrapper/wrap_mpi.cpp", 
                        ],
                    include_dirs=INCLUDE_DIRS + EXTRA_INCLUDE_DIRS,
                    library_dirs=LIBRARY_DIRS + EXTRA_LIBRARY_DIRS,
                    libraries=LIBRARIES + EXTRA_LIBRARIES,
                    define_macros=list(EXTRA_DEFINES.iteritems()),
                    extra_compile_args=conf["CXXFLAGS"],
                    extra_link_args=conf["CXXFLAGS"],
                    ),
                ],
            data_files=[("include/hedge", glob.glob("src/cpp/*.hpp"))],
            )




if __name__ == '__main__':
    main()



# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




from __future__ import division
import numpy
import numpy.linalg as la




def main() :
    from hedge.timestep import RK4TimeStepper
    from hedge.discretization import Discretization, ones_on_boundary, integral
    from hedge.visualization import SiloVisualizer, VtkVisualizer
    from hedge.tools import mem_checkpoint
    from pytools.stopwatch import Job
    from math import sin, cos, pi, sqrt
    from hedge.parallel import \
            guess_parallelization_context, \
            reassemble_volume_field
    from math import floor

    def f(x):
        return sin(pi*x)
        #if int(floor(x)) % 2 == 0:
            #return 1
        #else:
            #return 0

    def u_analytic(x, t):
        return f((-v*x/norm_v+t*norm_v))

    def boundary_tagger(vertices, el, face_nr):
        if numpy.dot(el.face_normals[face_nr], v) < 0:
            return ["inflow"]
        else:
            return ["outflow"]

    pcon = guess_parallelization_context()

    dim = 1

    job = Job("mesh")
    if dim == 1:
        v = numpy.array([1])
        if pcon.is_head_rank:
            from hedge.mesh import make_uniform_1d_mesh
            mesh = make_uniform_1d_mesh(-2, 5, 10, periodic=True)

        from hedge.element import IntervalElement
        el_class = IntervalElement

    elif dim == 2:
        v = numpy.array([2,0])
        if pcon.is_head_rank:
            from hedge.mesh import \
                    make_disk_mesh, \
                    make_square_mesh, \
                    make_rect_mesh, \
                    make_regular_square_mesh, \
                    make_regular_rect_mesh, \
                    make_single_element_mesh
        
            #mesh = make_square_mesh(max_area=0.0003, boundary_tagger=boundary_tagger)
            #mesh = make_regular_square_mesh(a=-r, b=r, boundary_tagger=boundary_tagger, n=3)
            #mesh = make_single_element_mesh(boundary_tagger=boundary_tagger)
            #mesh = make_disk_mesh(r=pi, boundary_tagger=boundary_tagger, max_area=0.5)
            #mesh = make_disk_mesh(boundary_tagger=boundary_tagger)
            
            if False:
                mesh = make_regular_rect_mesh(
                        (-0.5, -1.5),
                        (5, 1.5),
                        n=(10,5),
                        boundary_tagger=boundary_tagger,
                        periodicity=(True, False),
                        )
            if True:
                mesh = make_rect_mesh(
                        (-1, -1.5),
                        (5, 1.5),
                        max_area=0.3,
                        boundary_tagger=boundary_tagger,
                        periodicity=(True, False),
                        subdivisions=(10,5),
                        )

        from hedge.element import TriangularElement
        el_class = TriangularElement
    elif dim == 3:
        v = numpy.array([0,0,0.3])
        if pcon.is_head_rank:
            from hedge.mesh import make_cylinder_mesh, make_ball_mesh, make_box_mesh

            mesh = make_cylinder_mesh(max_volume=0.004, boundary_tagger=boundary_tagger,
                    periodic=False, radial_subdivisions=32)
            #mesh = make_box_mesh(dimensions=(1,1,2*pi/3), max_volume=0.01,
                    #boundary_tagger=boundary_tagger)
            #mesh = make_box_mesh(max_volume=0.01, boundary_tagger=boundary_tagger)
            #mesh = make_ball_mesh(boundary_tagger=boundary_tagger)
            #mesh = make_cylinder_mesh(max_volume=0.01, boundary_tagger=boundary_tagger)

        from hedge.element import TetrahedralElement
        el_class = TetrahedralElement
    else:
        raise RuntimeError, "bad number of dimensions"

    norm_v = la.norm(v)

    if pcon.is_head_rank:
        mesh_data = pcon.distribute_mesh(mesh)
    else:
        mesh_data = pcon.receive_mesh()
    job.done()

    job = Job("discretization")
    #mesh_data = mesh_data.reordered_by("cuthill")
    discr = pcon.make_discretization(mesh_data, el_class(8))
    vis_discr = discr
    job.done()

    vis = SiloVisualizer(vis_discr, pcon)
    #vis = VtkVisualizer(vis_discr, pcon, "fld")

    # operator setup ----------------------------------------------------------
    from hedge.data import \
            ConstantGivenFunction, \
            TimeConstantGivenFunction, \
            TimeDependentGivenFunction
    from hedge.operators import StrongAdvectionOperator, WeakAdvectionOperator
    op = WeakAdvectionOperator(discr, v, 
            inflow_u=TimeConstantGivenFunction(ConstantGivenFunction()),
            #inflow_u=TimeDependentGivenFunction(u_analytic)),
            flux_type="upwind")

    #from pyrticle._internal import ShapeFunction
    #sf = ShapeFunction(1, 2, alpha=1)

    def gauss_hump(x):
        from math import exp
        rsquared = numpy.dot(x, x)/(0.3**2)
        return exp(-rsquared)
    def gauss2_hump(x):
        from math import exp
        rsquared = (x*x)/(0.1**2)
        return exp(-rsquared)-0.5*exp(-rsquared/2)

    hump_width = 2
    def c_inf_hump(x):
        if abs(x) > hump_width:
            return 0
        else:
            exp(-1/(x-hump_width)**2)* exp(-1/(x+hump_width)**2)

    #u = discr.interpolate_volume_function(lambda x: u_analytic(x, 0))
    u = discr.interpolate_volume_function(gauss_hump)
    u /= integral(discr, u)

    # timestep setup ----------------------------------------------------------
    stepper = RK4TimeStepper()

    dt = discr.dt_factor(op.max_eigenvalue())
    nsteps = int(700/dt)

    if pcon.is_head_rank:
        print "%d elements, dt=%g, nsteps=%d" % (
                len(discr.mesh.elements),
                dt,
                nsteps)

    # diagnostics setup -------------------------------------------------------
    from pytools.log import LogManager, \
            add_general_quantities, \
            add_simulation_quantities, \
            add_run_info

    logmgr = LogManager("advection.dat", "w", pcon.communicator)
    add_run_info(logmgr)
    add_general_quantities(logmgr)
    add_simulation_quantities(logmgr, dt)
    discr.add_instrumentation(logmgr)

    from pytools.log import IntervalTimer
    vis_timer = IntervalTimer("t_vis", "Time spent visualizing")
    logmgr.add_quantity(vis_timer)
    stepper.add_instrumentation(logmgr)

    from hedge.log import Integral, LpNorm
    u_getter = lambda: u
    logmgr.add_quantity(Integral(u_getter, discr, name="int_u"))
    logmgr.add_quantity(LpNorm(u_getter, discr, p=1, name="l1_u"))
    logmgr.add_quantity(LpNorm(u_getter, discr, name="l2_u"))

    logmgr.add_watches(["step.max", "t_sim.max", "l2_u", "t_step.max"])

    # timestep loop -----------------------------------------------------------
    def logmap(x, low_exp=15):
        return 0.1*numpy.log10(numpy.abs(x)+1e-15)

    from hedge.discretization import Filter, ExponentialFilterResponseFunction
    filter = Filter(discr, ExponentialFilterResponseFunction(0.97, 3))

    for step in xrange(nsteps):
        logmgr.tick()

        t = step*dt

        if step % 5 == 0:
            vis_timer.start()
            visf = vis.make_file("fld-%04d" % step)
            vis.add_data(visf, [
                        ("u", u), 
                        ("logu", logmap(u)), 
                        #("u_true", u_true), 
                        ], 
                        #expressions=[("error", "u-u_true")]
                        time=t, 
                        step=step
                        )
            visf.close()
            vis_timer.stop()


        #u = filter(stepper(u, t, dt, op.rhs))
        u = stepper(u, t, dt, op.rhs)
        #if step % 1 == 0:
            #u = filter(u)

        #u_true = discr.interpolate_volume_function(
                #lambda x: u_analytic(t, x))

    vis.close()

    logmgr.tick()
    logmgr.save()


if __name__ == "__main__":
    #import cProfile as profile
    #profile.run("main()", "advec.prof")
    main()


